import numpy as np
from collections import OrderedDict


class PointsGenerator:
    """Class that generates a dictionary of operation points for an emva test.

    The points are stored in the :attr:`points` attribute as a dictionary.
    The points are ordered if they are a 'spatial' or 'temporal' operation
    point. Under each keys there is a dictionary whose keys are the
    different exposure times and their values are the list of radiances
    under which the camera is illuminated.
    """
    def __init__(self,
                 cam,
                 exposure_min=None,
                 exposure_max=None,
                 exposure_fixed=None,
                 radiance_min=None,
                 radiance_max=None,
                 gain=None,
                 blackref=None,
                 steps=100):
        """Point generator init method.

        Parameters
        ----------
        cam : The camera object that will be taking the images.
        exposure_min : float, optional
                       The minimal exposure time (in ns).
        exposure_max : float, optional
                       The maximal exposure time (in ns).
        exposure_fixed : float, optional
                         By default, the points given are for an exposure time
                         variation test (if this is None). If a value is given
                         to this kwarg, this will be the camera's exposure
                         time (in ns) at which the operation points will be
                         set for an illumination variation test.
        radiance_min : float, optional
                       The minimal radiance (in W/cm^2/sr). If None, a value
                       above dark illumination will be automatically chosen.
        radiance_max : float, optional
                       The maximal radiance (in W/cm^2/sr).
                       If None, the maximal
                       radiation will be taken as the saturation radiation for
                       the exposition time given in the exposure_fixed kwarg.
        gain : float, optional
               The camera's gain at which we want the test to run.
        blackref : float, optional
                   The camera's blackoffset at which we want the test to run.
        steps : int, optional
                The number of points in the test.
        """
        self._cam = cam
        self._steps = steps
        self._exposure_min = exposure_min or self._cam.exposure_min
        self._exposure_max = exposure_max or self._cam.exposure_max
        self._exposure = exposure_fixed
        self._radiance_min = radiance_min
        self._radiance_max = radiance_max
        self._gain = gain or self._cam.K
        self._blackref = blackref or self._cam.blackoffset

        if self._exposure is None:
            # Get radiance for saturation at maximal exposure time
            # Only if it is for an exposure time variation test
            self._cam.exposure = self._exposure_max
            self._radiance = self._cam.get_radiance_for()

        else:
            # get radiances for radiation variation
            self._cam.exposure = self._exposure
            self._cam.K = self._gain
            self._cam.blackoffset = self._blackref
            m = self._cam.grab(0.0).mean()
            target = (self._cam.img_max - m) / self._steps + m
            self._radiance_min = self._cam.get_radiance_for(mean=target)
            self._radiance_max = self._cam.get_radiance_for()

        # By default, an exposure time variation data points
        # If exposure fixed is given, it is photons variation
        self._points = self._get_points()

    def _get_points(self):
        spatial = OrderedDict()
        temporal = OrderedDict()

        if self._exposure is None:
            # Exposure time variation
            # round to only have one decimal
            exposures = np.round(np.linspace(self._exposure_min,
                                             self._exposure_max,
                                             self._steps), 1)
            # only one radiance
            radiances = [self._radiance, 0.0]

            # Main loop to compute points
            for n, texp in enumerate(exposures):
                if self._is_point_spatial(n):
                    spatial[texp] = radiances
                temporal[texp] = radiances

        else:
            # Photons variations
            # only one exposure time
            radiances = np.linspace(self._radiance_min,
                                    self._radiance_max,
                                    self._steps).tolist()
            # round to only have one decimal
            self._exposure = round(self._exposure)
            radiances.append(0.0)
            spatial[self._exposure] = [radiances[self._steps // 2], 0.0]
            temporal[self._exposure] = radiances

        return {'spatial': spatial, 'temporal': temporal}

    @property
    def points(self):
        """The operation points."""
        return self._points

    def _is_point_spatial(self, n):
        """Checks if a spatial test must be done at this point.

        Spatial test are executed at mid run for both dark and bright tests.
        """
        middle = self._steps // 2
        if n in (middle, self._steps + middle):
            return True
        return False



# -*- coding: utf-8 -*-
# Copyright (c) 2014 The EMVA1288 Authors. All rights reserved.
# Use of this source code is governed by a GNU GENERAL PUBLIC LICENSE that can
# be found in the LICENSE file.

"""Compute EMVA1288 values from data
This class takes the data from data.Data1288 and compute the actual EMVA1288
values.
"""

from __future__ import print_function
import logging
import numpy as np
from emva1288.process import routines
from scipy.ndimage import convolve
import numpy.ma as ma


class Results1288(object):
    """Class used to process data and to generate pdf report using LaTeX.

    When properties from this class are computed, their docstring are also
    parsed using the :func:`~emva1288.process.routines.cls_1288_info`
    function to retrieve more data informations (like units or full name).
    """
    #########################################################################
    # Docstrings with .. emva1288:: flags are used when computing the       #
    # properties to add relevant information to data for further processing #
    # like plotting or creating a report                                    #
    #########################################################################
    def __init__(self,
                 data,
                 pixel_area=None,
                 index_u_ysat=None,
                 loglevel=logging.DEBUG):
        """Results computation init method.

        This class uses a :class:`python:logging.Logger` object to display
        informations for users.

        Parameters
        ----------
        data : dict
               The data dictionary to compute the results from.
        pixel_area : float, optional
                     The area of one pixel in um^2.
        index_u_ysat : int, optional
                       The index of the u_y array at which we consider
                       that the camera saturates. This is used if forcing
                       the saturation point is necessary.
        loglevel : int, optional
                   The level for the :class:`python:logging.Logger` object.
        """
        logging.basicConfig()
        self.log = logging.getLogger('Results')
        self.log.setLevel(loglevel)

        self.temporal = data['temporal']
        self.spatial = data['spatial']

        self.pixel_area = pixel_area
        self._s2q = 1.0 / 12.0
        self._index_start = 0
        self._index_sensitivity_min = 0

        self._histogram_Qmax = 256  # Maximum number of bins in histograms

        # Sometimes we need to force the saturation point
        # in those cases pass the index in the initialization of Results1288
        self._index_u_ysat = index_u_ysat

    @property
    def s2q(self):
        """Quantification noise."""
        return self._s2q

    @property
    def index_start(self):
        """The array's starting index.

        .. emva1288::
            :Section: info
            :Short: Start array index
        """
        return self._index_start

    @property
    def index_u_ysat(self):
        """Index of saturation.

        .. emva1288::
            :Section: sensitivity
            :Short: Saturation index
        """

        if self._index_u_ysat:
            return self._index_u_ysat

        max_ = 0
        max_i = 0

        # we have to loop backwards because sometimes we have some
        # noise pics that really bother the computation
        s2_y = self.temporal['s2_y']
        for i in range(len(s2_y) - 1, -1, -1):
            # Check that is not a local max
            if (s2_y[i] >= max_) or \
               (s2_y[abs(i - 1)] >= max_):
                max_ = s2_y[i]
                max_i = i
            elif (s2_y[i] < max_) and \
                 (s2_y[abs(i - 1)] < max_):
                break

        self._index_u_ysat = max_i
        return self._index_u_ysat

    @property
    def index_sensitivity_max(self):
        """Index for linear fits in sensitivity part of the standard
        (70% of saturation).

        .. emva1288::
            :Section: sensitivity
            :Short: Sensitivity fit maximum index
        """

        Y = self.temporal['u_y'] - self.temporal['u_ydark']
        m = 0.7 * (Y[self.index_u_ysat])
        return max(np.argwhere(Y <= m))[0]

    @property
    def index_sensitivity_min(self):
        """Sensitivity minimum index.

        Index for linear fits in sensitivity part of the standard
        (70% of saturation)

        .. emva1288::
            :Section: sensitivity
            :Short: Sensitivity fit minimum index
        """
        return self._index_sensitivity_min

    @property
    def R(self):
        """Responsivity.

        Slope of the (u_y - u_y_dark) Vs u_p. Fit with offset = 0
        Uses the :func:`~emva1288.process.routines.LinearB0` function
        to make the fit.

        .. emva1288::
            :Section: sensitivity
            :Short: Responsivity
            :Symbol: R
            :Unit: DN/p
        """

        Y = self.temporal['u_y'] - self.temporal['u_ydark']
        X = self.temporal['u_p']

        val, _error = routines.LinearB0(X[self.index_sensitivity_min:
                                          self.index_sensitivity_max + 1],
                                        Y[self.index_sensitivity_min:
                                          self.index_sensitivity_max + 1])

        return val[0]

    @property
    def K(self):
        """Overall system gain.

        Slope of (s2_y - s2_y_dark) Vs (u_y - u_y_dark). Fit with
        offset = 0. Uses the :func:`~emva1288.process.routines.LinearB0`
        to make the fit.

        .. emva1288::
            :Section: sensitivity
            :Short: System gain
            :Symbol: K
            :Unit: $DN/e^-$
            :LatexName: K
        """

        X = self.temporal['u_y'] - self.temporal['u_ydark']
        Y = self.temporal['s2_y'] - self.temporal['s2_ydark']

        val, _error = routines.LinearB0(X[self.index_sensitivity_min:
                                          self.index_sensitivity_max + 1],
                                        Y[self.index_sensitivity_min:
                                          self.index_sensitivity_max + 1])

        return val[0]

    def inverse_K(self):
        """Inverse of overall system gain.

        .. emva1288::
            :Section: sensitivity
            :Short: Inverse of overall system gain
            :Symbol: 1/K
            :Unit: $e^-/DN$
            :LatexName: InvK
        """

        return 1. / self.K

    @property
    def QE(self):
        """Quantum efficiency.

        It is retrieved as the ratio of the responsivity to the overall
        system gain.

        .. emva1288::
            :Section: sensitivity
            :Short: Quantum efficiency
            :Symbol: $\eta$
            :Unit: \%
            :LatexName: QE
        """

        return 100.0 * self.R / self.K

    @property
    def sigma_y_dark(self):
        """Temporal Dark Noise.

        Uses :func:`~emva1288.process.routines.LinearB` to make the fit.

        .. emva1288::
            :Section: sensitivity
            :Short: Temporal Dark Noise
            :Symbol: $\sigma_{y.dark}$
            :Unit: DN
            :LatexName: SigmaYDark
        """

        if len(np.unique(self.temporal['texp'])) <= 2:
            s2_ydark = self.temporal['s2_ydark'][0]
        else:
            fit, _error = routines.LinearB(self.temporal['texp'],
                                           self.temporal['s2_ydark'])
            s2_ydark = fit[1]

        # Lower limit for the temporal dark noise
        # The temporal dark noise in this range is dominated by the
        # quantization noise
        if s2_ydark < 0.24:
            s2_ydark = 0.24

        return np.sqrt(s2_ydark)

    @property
    def sigma_d(self):
        """Temporal Dark Noise.

        .. emva1288::
            :Section: sensitivity
            :Short: Temporal Dark Noise
            :Symbol: $\sigma_d$
            :Unit: $e^-$
            :LatexName: SigmaDark
        """

        return np.sqrt((self.sigma_y_dark ** 2) - self._s2q) / self.K

    @property
    def u_p_min(self):
        """Absolute sensitivity threshold.

        .. emva1288::
            :Section: sensitivity
            :Short: Absolute sensitivity threshold
            :Symbol: $\mu_{p.min}$
            :Unit: $p$
            :LatexName: UPMin
        """

        return (100.0 / self.QE) * ((self.sigma_y_dark / self.K) + 0.5)

    @property
    def u_p_min_area(self):
        """Sensitivity threshold per pixel area.

        Returns None if pixel area is not defined or 0.

        .. emva1288::
            :Section: sensitivity
            :Short: Sensitivity threshold
            :Symbol: $\mu_{p.min.area}$
            :Unit: $p/\mu m^2$
            :LatexName: UPMin
        """

        if not self.pixel_area:
            return None

        return self.u_p_min / self.pixel_area

    @property
    def u_e_min(self):
        """Sensitivity threshold.

        .. emva1288::
            :Section: sensitivity
            :Short: Sensitivity threshold
            :Symbol: $\mu_{e.min}$
            :Unit: $e^-$
            :LatexName: UEMin
        """
        return self.QE * self.u_p_min / 100.0

    @property
    def u_e_min_area(self):
        """Sensitivity threshold per pixel area.

        Returns None if the pixel area is not defined or 0.

        .. emva1288::
            :Section: sensitivity
            :Short: Sensitivity threshold
            :Symbol: $\mu_{e.min.area}$
            :Unit: $e^-/\mu m^2$
            :LatexName: UEMin
        """
        if not self.pixel_area:
            return None

        return self.u_e_min / self.pixel_area

    @property
    def u_p_sat(self):
        """Saturation Capacity.

        .. emva1288::
            :Section: sensitivity
            :Short: Saturation Capacity
            :Symbol: $\mu_{p.sat}$
            :Unit: $p$
            :LatexName: UPSat
        """

        return self.temporal['u_p'][self.index_u_ysat]

    @property
    def u_p_sat_area(self):
        """Saturation Capacity per pixel area.

        Returns None if pixel area is not defined or 0.

        .. emva1288::
            :Section: sensitivity
            :Short: Saturation Capacity
            :Symbol: $\mu_{p.sat.area}$
            :Unit: $p/\mu m^2$
            :LatexName: UPSat
        """

        if not self.pixel_area:
            return None

        return self.u_p_sat / self.pixel_area

    @property
    def u_e_sat(self):
        """Saturation Capacity.

        Number of electrons at saturation.

        .. emva1288::
            :Section: sensitivity
            :Short: Saturation Capacity
            :Symbol: $\mu_{e.sat}$
            :Unit: $e^-$
            :LatexName: UESat
        """

        return self.QE * self.u_p_sat / 100.0

    @property
    def u_e_sat_area(self):
        """Saturation Capacity per pixel area.

        Returns None if pixel area is not defined or 0.

        .. emva1288::
            :Section: sensitivity
            :Short: Saturation Capacity
            :Symbol: $\mu_{e.sat.area}$
            :Unit: $e^-/\mu m^2$
            :LatexName: UESat
        """

        if not self.pixel_area:
            return None

        return self.u_e_sat / self.pixel_area

    @property
    def SNR_max(self):
        """Maximum Signal-to-Noise Ratio.

        .. emva1288::
            :Section: sensitivity
            :Short: Signal-to-Noise Ratio
            :Symbol: $SNR_{max}$
            :LatexName: SNRMax
        """

        return np.sqrt(self.u_e_sat)

    def SNR_max_dB(self):
        """Maximum Signal to Noise Ratio in Db.

        .. emva1288::
            :Section: sensitivity
            :Short: Maximum Signal to Noise Ratio in Db
            :Symbol: $SNR_{max.dB}$
            :Unit: dB
            :LatexName: SNRMaxDB
        """

        return 20. * np.log10(self.SNR_max)

    def SNR_max_bit(self):
        """Maximum Signal to Noise Ratio in Bits.

        .. emva1288::
            :Section: sensitivity
            :Short: Maximum Signal to Noise Ratio in Bits
            :Symbol: $SNR_{max.bit}$
            :Unit: bit
            :LatexName: SNRMaxBit
        """

        return np.log2(self.SNR_max)

    def inverse_SNR_max(self):
        """Inverse Maximum Signal to Noise Ratio.

        .. emva1288::
            :Section: sensitivity
            :Short: Inverse Maximum Signal to Noise Ratio
            :Symbol: $SNR_{max}^{-1}$
            :Unit: \%
            :LatexName: InvSNRMax
        """

        return 100. / self.SNR_max

    @property
    def DR(self):
        """Dynamic Range.

        Defined as the saturation capacity devided by the absolute sensitivity
        threshold. The greater this number is, the greater the operational
        range of a camera (between the dark noise level and the saturation
        level).

        .. emva1288::
            :Section: sensitivity
            :Short: Dynamic Range
            :Symbol: DR
            :LatexName: DR
        """

        return self.u_p_sat / self.u_p_min

    def DR_dB(self):
        """Dynamic Range in deciBels.

        It is defined as 20 * log_10 ( Dynamic Range ).

        .. emva1288::
            :Section: sensitivity
            :Short: Dynamic Range
            :Symbol: $DR_{dB}$
            :Unit: dB
            :LatexName: DRDB
        """

        return 20. * np.log10(self.DR)

    def DR_bit(self):
        """Dynamic Range in bits.

        It is defined as log_2 ( Dynamic Range ).

        .. emva1288::
            :Section: sensitivity
            :Short: Dynamic Range
            :Symbol: $DR_{bit}$
            :Unit: bit
            :LatexName: DRBit
        """

        return np.log2(self.DR)

    @property
    def index_linearity_min(self):
        """Linearity fit minimun index.

        Minimum index for linear fit (5% of saturation).

        .. emva1288::
            :Section: linearity
            :Short: Linearity fit minimun index
        """
        Y = self.temporal['u_y'] - self.temporal['u_ydark']
        vmin = 0.05 * (Y[self.index_u_ysat])

        return min(np.argwhere(Y >= vmin))[0]

    @property
    def index_linearity_max(self):
        """Linearity fit maximum index.

        Maximum index for linear fit (95% of saturation).

        .. emva1288::
            :Section: linearity
            :Short: Linearity fit maximum index
        """

        Y = self.temporal['u_y'] - self.temporal['u_ydark']
        vmax = 0.95 * (Y[self.index_u_ysat])
        return max(np.argwhere(Y <= vmax))[0]

    def linearity(self):
        """Returns a dictionary containing linearity information.

        It fits the mean digital signal in function of the mean photon count
        (Linear fit) using the EMVA1288 standard for linear fit.

        Returns
        -------
        dict : Linearity dictionary.
               The keys are:

               - *'fit_slope'* : The slope of the linear fit.
               - *'fit_offset'* : The offset of the fit.
               - *'relative_deviation'* : The relative deviation of the real
                 data from the fit (in %) for the whole array.
               - *'linearity_error_min'* : The minimal value of the
                 relative deviation.
               - *'linearity_error_max'* : The maximal value of the
                 relative deviation.
        """
        Y = self.temporal['u_y'] - self.temporal['u_ydark']
        X = self.temporal['u_p']

        # The maximum index has to be included, this is the reason for the +1
        imax = self.index_linearity_max + 1
        imin = self.index_linearity_min

        ##################################################################
        # Following the emva1288 standart for the computation of the fit #
        ##################################################################

        X_ = X[imin: imax]
        Y_ = Y[imin: imax]

        xy = np.sum(X_ / Y_)
        xy2 = np.sum(X_ / (Y_ ** 2))
        x2y2 = np.sum((X_ / Y_) ** 2)
        _y = np.sum(1. / Y_)
        _y2 = np.sum(1. / (Y_ ** 2))

        b = ((xy * xy2) - (x2y2 * _y)) / ((xy2 ** 2) - (x2y2 * _y2))
        a = (xy - (b * xy2)) / x2y2

        # The equivalent numpy polyfit is
        # a, b = np.polynomial.polynomial.polyfit(X_, Y_, 1, w=1 / Y_)

        dev = 100. * (Y - (a * X + b)) / (a * X + b)

        lin = {}
        lin['fit_slope'] = a
        lin['fit_offset'] = b
        lin['relative_deviation'] = dev
        lin['linearity_error_min'] = np.min(dev[imin: imax])
        lin['linearity_error_max'] = np.max(dev[imin: imax])

        return lin

    @property
    def LE_min(self):
        """ Min Linearity error.

        .. emva1288::
            :Section: linearity
            :Short: Min Linearity error
            :Symbol: $LE_{min}$
            :Unit: \%
            :LatexName: LEMin
        """

        return self.linearity()['linearity_error_min']

    @property
    def LE_max(self):
        """Max Linearity error.

        .. emva1288::
            :Section: linearity
            :Short: Max Linearity error
            :Symbol: $LE_{max}$
            :Unit: \%
            :LatexName: LEMax
        """

        return self.linearity()['linearity_error_max']

    @property
    def u_I_var_DN(self):
        """Dark Current from variance.

        The dark current from variance is the square root of the slope of the
        dark signal variance in function of the exposure time.
        Returns NaN if u_I_var is imaginary (if the fit slope is negative).

        .. emva1288::
            :Section: dark_current
            :Short: Dark Current from variance
            :Symbol: $\mu_{I.var.DN}$
            :Unit: $DN/s$
            :LatexName: UIVar
        """
        if len(np.unique(self.temporal['texp'])) <= 2:
            return np.nan

        fit, _error = routines.LinearB(self.temporal['texp'],
                                       self.temporal['s2_ydark'])

        if fit[0] < 0:
            return np.nan
        # Multiply by 10^9 because exposure times are in nanoseconds
        return fit[0] * (10 ** 9)

    @property
    def u_I_var(self):
        """Dark Current from variance.

        The dark current from variance is the square root of the slope of the
        dark signal variance in function of the exposure times divided
        by the overall system gain.
        Returns NaN if u_I_var is imaginary (if the fit slope is negative).

        .. emva1288::
            :Section: dark_current
            :Short: Dark Current from variance
            :Symbol: $\mu_{I.var}$
            :Unit: $e^-/s$
            :LatexName: UIVar
        """
        ui = self.u_I_var_DN
        if ui is np.nan:
            return np.nan

        return ui / (self.K ** 2)

    @property
    def u_I_mean_DN(self):
        """Dark Current from mean.

        The dark current from mean is the slope of the dark signal mean in
        function of the exposure time.
        Returns NaN if the number of different exposure times is less than 3.

        .. emva1288::
            :Section: dark_current
            :Short: Dark Current from mean
            :Symbol: $\mu_{I.mean.DN}$
            :Unit: $DN/s$
        """

        if len(np.unique(self.temporal['texp'])) <= 2:
            return np.nan

        fit, _error = routines.LinearB(self.temporal['texp'],
                                       self.temporal['u_ydark'])
        # Multiply by 10 ^ 9 because exposure time in nanoseconds
        return fit[0] * (10 ** 9)

    @property
    def u_I_mean(self):
        """Dark Current from mean.

        The dark current from mean is the slope of the dark signal mean in
        function of the exposure times divided by the overall system gain.
        Returns NaN if the number of different exposure times is less than 3.

        .. emva1288::
            :Section: dark_current
            :Short: Dark Current from mean
            :Symbol: $\mu_{I.mean}$
            :Unit: $e^-/s$
        """
        ui = self.u_I_mean_DN
        if ui is np.nan:
            return np.nan
        return ui / self.K

    @property
    def sigma_2_y_stack(self):
        """Temporal variance stack.

        Mean value of the bright variance image.

        .. emva1288::
            :Section: spatial
            :Short: Temporal variance stack
            :Symbol: $\sigma^2_{y.stack}$
            :Unit: DN2
        """

        return self.spatial['var_mean']

    @property
    def sigma_2_y_stack_dark(self):
        """Temporal variance stack dark.

        Mean value of the dark variance image.

        .. emva1288::
            :Section: spatial
            :Short: Temporal variance stack dark
            :Symbol: $\sigma^2_{y.stack.dark}$
            :Unit: DN2
        """

        return self.spatial['var_mean_dark']

    @property
    def s_2_y_measured(self):
        """Spatial variance measure.

        Variance value of the bright variance image.

        .. emva1288::
            :Section: spatial
            :Short: Spatial variance measure
            :Symbol: $s^2_{y.measured}$
            :Unit: DN2
        """
        # ddof = 1 (delta degrees of freedom) accounts for the minus 1 in the
        # divisor for the calculation of variance
        return self.spatial['avg_var']

    @property
    def s_2_y_measured_dark(self):
        """Spatial variance measured dark.

        Variance value of the dark variance image.

        .. emva1288::
            :Section: spatial
            :Short: Spatial variance measured dark
            :Symbol: $s^2_{y.measured.dark}$
            :Unit: DN2
        """
        return self.spatial['avg_var_dark']

    @property
    def s_2_y(self):
        """Spatial variance from image.

        .. emva1288::
            :Section: spatial
            :Short: Spatial variance from image
            :Symbol: $s^2_{y}$
            :Unit: DN2
        """
        return self.s_2_y_measured - (self.sigma_2_y_stack /
                                      self.spatial['L'])

    @property
    def s_2_y_dark(self):
        """Spatial variance from image,

        .. emva1288::
            :Section: spatial
            :Short: Spatial variance from image
            :Symbol: $s^2_{y}$
            :Unit: DN2
        """
        return self.s_2_y_measured_dark - (self.sigma_2_y_stack_dark /
                                           self.spatial['L_dark'])

    @property
    def DSNU1288(self):
        """DSNU.

        Dark Signal NonUniformity (in e^-) is defined as the deviation
        standard of the dark signal devided by the overall system gain.
        If the variance is negative, it will return NaN instead of an
        imaginary number.

        .. emva1288::
            :Section: spatial
            :Short: DSNU
            :Symbol: $DSNU_{1288}$
            :Unit: $e^-$
            :LatexName: DSNU
        """

        if self.s_2_y_dark < 0.:
            return np.nan
        return np.sqrt(self.s_2_y_dark) / self.K

    def DSNU1288_DN(self):
        """DSNU in DN.

        Defined as the DSNU in e^- multiplied by the overall system gain.
        Returns NaN if the dark signal variance is negative.

        Returns
        -------
        float : The DSNU in DN.


        .. emva1288::
            :Section: spatial
            :Short: DSNU in DN
            :Symbol: $DSNU_{1288.DN}$
            :Unit: DN
            :LatexName: DSNUDN
        """

        if self.s_2_y_dark < 0:
            return np.nan
        return np.sqrt(self.s_2_y_dark)

    @property
    def PRNU1288(self):
        """PRNU.

        Photo Response NonUniformity (in %) is defined as the square root of
        the difference between the spatial variance of a bright image (or from
        an average of bright images to remove temporal difformities) and the
        spatial variance of dark signal, divided by the difference between the
        mean of a bright image and the mean of a dark image.

        .. emva1288::
            :Section: spatial
            :Short: PRNU
            :Symbol: $PRNU_{1288}$
            :Unit: \%
            :LatexName: PRNU
        """

        return (np.sqrt(self.s_2_y - self.s_2_y_dark) * 100 /
                (self.spatial['avg_mean'] - self.spatial['avg_mean_dark']))

    @property
    def histogram_PRNU(self):
        """PRNU histogram.

        Uses the :func:`~emva1288.process.routines.high_pass_filter` function
        to filter the image and then uses the
        :func:`~emva1288.process.routines.Histogram1288` function
        to make the histogram.

        .. emva1288::
            :Section: defect_pixel
            :Short: PRNU histogram
        """

        # For prnu, perform the convolution
        y = self.spatial['sum'] - self.spatial['sum_dark']
        # Applying the filter on the image
        data = routines.high_pass_filter(y, 5)
        y = data['img']

        h = routines.Histogram1288(y, self._histogram_Qmax)
        # Rescale the bins
        h['bins'] /= (self.spatial['L'] * data['multiplicator'])

        return h

    @property
    def histogram_PRNU_accumulated(self):
        """Accumulated PRNU histogram.

        Uses the :func:`~emva1288.process.routines.high_pass_filter` function
        to filter the image and then uses the
        :func:`~emva1288.process.routines.Histogram1288` function
        to make the histogram.

        .. emva1288::
            :Section: defect_pixel
            :Short: accumulated PRNU histogram
        """

        # For prnu, perform the convolution
        y = self.spatial['sum'] - self.spatial['sum_dark']
        data = routines.high_pass_filter(y, 5)
        y = data['img']
        # For the accumulated histogram substract the mean
        y = np.abs(y - int(np.mean(y)))

        h = routines.Histogram1288(y, self._histogram_Qmax)
        # Rescale the bins
        h['bins'] /= (self.spatial['L'] * data['multiplicator'])

        # Perform the cumulative summation
        h['values'] = np.cumsum(h['values'][::-1])[::-1]
        # we want it as percentage of pixels
        h['values'] = 100 * h['values'] / y.size

        return h

    @property
    def histogram_DSNU(self):
        """DSNU histogram.

        Uses the :func:`~emva1288.process.routines.Histogram1288` function
        to make the histogram.

        .. emva1288::
            :Section: defect_pixel
            :Short: DSNU histogram
        """

        # For dsnu, the image is just the dark image, upscaled to have
        # only integers
        y = self.spatial['sum_dark']

        h = routines.Histogram1288(y, self._histogram_Qmax)
        # Rescale the bins, this is due to upscaling the average image to have
        # only integers
        scale = self.spatial['L_dark'] * 25.
        h['bins'] /= scale
        # The image is not centered around zero, so we shift the bins
        h['bins'] -= (y.mean() / scale)
        return h

    @property
    def histogram_DSNU_accumulated(self):
        """Accumulated DSNU histogram.

        Uses the :func:`~emva1288.process.routines.Histogram1288` function
        to make the histogram.

        .. emva1288::
            :Section: defect_pixel
            :Short: accumulated DSNU histogram
        """

        # Dark image upscaled to have only integers
        y = self.spatial['sum_dark']
        # For the accumulated dsnu histogram, substract the mean from the image
        y = np.abs(y - int(np.mean(y)))

        h = routines.Histogram1288(y, self._histogram_Qmax)
        # Rescale the bins
        h['bins'] /= (self.spatial['L_dark'] * 25.)

        # Perform the cumulative summation (the ::-1 means backwards)
        # because the cumsum function is performed contrary to what we need
        h['values'] = np.cumsum(h['values'][::-1])[::-1]

        # we want it as percentage of pixels
        h['values'] = 100 * h['values'] / y.size

        return h

    def xml(self, filename=None):  # pragma: no cover
        """Method that writes the results in xml format to a file.

        Parameters
        ----------
        filename : str, optional
                   The file to write the results. If None, the xml string
                   won't be written but will be returned instead.

        Returns
        -------
        str : If the xml string is not written into a file, it is returned.
        """
        results = self.results_by_section
        if not filename:
            return routines.dict_to_xml(results)
        routines.dict_to_xml(results, filename=filename)

    @property
    def results(self):
        """Dictionnary with all the values and metadata for EMVA1288 values.

        It uses the :func:`~emva1288.process.routines.obj_to_dict` to compute
        all the results at once.
        """
        return routines.obj_to_dict(self)

    @property
    def results_by_section(self):  # pragma: no cover
        """Results ordered by section."""
        return routines._sections_first(self.results)

    def print_results(self):  # pragma: no cover
        """Print results to the screen."""
        results = self.results_by_section

        for section, attributes in results.items():
            print('*' * 50)
            print(section)
            print('-' * 50)

            for attribute, info in attributes.items():
                if 'value' not in info:
                    continue

                print('{:<50}{:<30}{:>10}'.format(info.get('short'),
                                                  str(info.get('symbol')),
                                                  str(info.get('value'))))
        print('*' * 50)
        print(' ')



# -*- coding: utf-8 -*-
# Copyright (c) 2014 The EMVA1288 Authors. All rights reserved.
# Use of this source code is governed by a GNU GENERAL PUBLIC LICENSE that can
# be found in the LICENSE file.

"""Utils functions

"""
from __future__ import print_function
import numpy as np
import os
from scipy.optimize import leastsq
from lxml import etree
from PIL import Image
from collections import OrderedDict
from emva1288.camera import routines
import numpy.ma as ma
from scipy.ndimage import convolve
import warnings
# import cv2


SIGNIFICANT_DIGITS = 7


def load_image(fname):
    img = Image.open(fname)
    img = np.asarray(img.split()[0])
    # img = cv2.imread(fname, cv2.CV_LOAD_IMAGE_UNCHANGED)
    # img = cv2.split(img)[0]
    return img


def get_int_imgs(imgs):
    '''
    Returns the sum and pseudo-variance from list of images
    sum is just the image resulting on the addition of all the images
    pvar is the pseudo-variance, this is
    pvar = SUM((Li - SUM(i))^2)
    to get variance from pseudo-variance
    var = (1/(L^2) * 1/(L - 1)) * pvar
    '''
    L = len(imgs)

    sum_ = 0
    sq_ = 0
    for img in imgs:
        # we force the images as int64 to make sure we do not clip
        i = img.astype(np.int64)
        sum_ += i
        sq_ += np.square(i)

    # the pseudo variance can be computed from the sum image and the sum of
    # the square images
    var_ = L * (L * sq_ - np.square(sum_))

    return {'L': L, 'sum': sum_, 'pvar': var_}


def LinearB0(Xi, Yi):
    X = np.asfarray(Xi)
    Y = np.asfarray(Yi)

    # we want a function y = m * x
    def fp(v, x):
        return x * v[0]

    # the error of the function e = x - y
    def e(v, x, y):
        return (fp(v, x) - y)

    # the initial value of m, we choose 1, because we thought YODA would
    # have chosen 1
    v0 = [1.0]

    vr, _success = leastsq(e, v0, args=(X, Y))

    # compute the R**2 (sqrt of the mean of the squares of the errors)
    err = np.sqrt(sum(np.square(e([vr], X, Y))) / (len(X) * len(X)))

    # Some versions of leastsq returns an array, other a scalar, so here we
    # make sure
    # it is an array
    val = np.array([vr]).flatten()

    return val, err


def LinearB(Xi, Yi):
    X = np.asfarray(Xi)
    Y = np.asfarray(Yi)

    # we want a function y = m * x + b
    def fp(v, x):
        return x * v[0] + v[1]

    # the error of the function e = x - y
    def e(v, x, y):
        return (fp(v, x) - y)

    # the initial value of m, we choose 1, because we thought YODA would
    # have chosen 1
    v0 = np.array([1.0, 1.0])

    vr, _success = leastsq(e, v0, args=(X, Y))

    # compute the R**2 (sqrt of the mean of the squares of the errors)
    err = np.sqrt(sum(np.square(e(vr, X, Y))) / (len(X) * len(X)))

#    print vr, success, err
    return vr, err


def GetImgShape(img):
    rows = 1

    if img.ndim == 1:
        cols, = img.shape
    else:
        rows, cols = img.shape
    return rows, cols


def FFT1288(img_, rotate=False, n=1):
    '''Compute the FFT emva1288 style

    Compute an FFT per line and average the resulting ffts

    Parameters
    ----------
    img : array
        Input image
    rotate : bool (optional)
        Rotate the image before performing the FFT
    n : int (optional)
        If the image is the sum of several images use this value
        to produce the fft of the average image

    Returns
    -------
    array : One dimension FFT power spectrum
    '''
    img = np.ma.copy(img_)
    if rotate:
        img = img.transpose()

    _rows, cols = GetImgShape(img)

    if isinstance(img, np.ma.masked_array):
        img_line = []
        # If img is masked, perform fft line by line
        for line in img:
            # fft must be performed on compressed line
            line = line.compressed()
            # Assuring the line is not empty
            if line.size == 0:
                continue
            img_line.append(line)
        # cols might be changed, it corresponds to len(arr)
        cols = np.min([len(line)for line in img_line])
        img = img_line[:][0:cols]

    # simply perform fft on x axis
    img = np.asfarray(img) - np.mean(img)
    fft = np.fft.fft(img, axis=1)
    fft /= np.sqrt(cols)

    fabs = np.real(fft * np.conjugate(fft))

    # extract the mean of each column of the fft
    r = np.mean(fabs, axis=0)

    # if the image is the sum of n image, the fft of the average image is
    r /= n ** 2

    # Return only half of the spectrogram (it is symemtrical)
    return r[: cols // 2]


def GetFrecs(fft):
    n = len(fft)
    x = np.arange(n)
    x = x * 1.0 / (2 * n)
    return x


def Histogram1288(img, Qmax):
    y = np.ravel(img)

    ymin = np.min(y)
    ymax = np.max(y)

    # Because we are working with integers, minimum binwidth is 1
    W = 1
    q = ymax - ymin
    Q = q + 1

    # When too many bins, create a new integer binwidth
    if Q > Qmax:
        # We want the number of bins as close as possible to Qmax (256)
        W = int(np.ceil(1. * q / (Qmax - 1)))
        Q = int(np.floor(1. * q / W)) + 1

    # The bins
    # we need one more value for the numpy histogram computation
    # numpy used bin limits
    # in our interpretation we use the lower limit of the bin
    B = [ymin + (i * W) for i in range(Q + 1)]

    # Normal distribution with the original sigma, and mean
    mu = np.mean(y)
    sigma = np.std(y)
    # For masked arrays, the total number of pixel is the sum of the non-masked
    # value
    N = np.ma.count(y)
    normal = ((1. * (ymax - ymin) / Q) *
              N / (np.sqrt(2 * np.pi) * sigma) *
              np.exp(-0.5 * (1. / sigma * (B[:-1] - mu)) ** 2))


#############################################
#
#   # Reference algorithm, it's pretty slow
#   # the numpy version gives the same results
#
#     #The histogram container
#     H = np.zeros((Q,), dtype=np.int64)
#
#     #Histogram computation
#     for yi in y:
#         q = (yi - ymin) / W
#         H[q] += 1
#############################################

    H, _b = np.histogram(y, B, range=(ymin, ymax))

    return {'bins': np.asfarray(B[:-1]), 'values': H, 'model': normal}


def cls_1288_info(cls):
    """Dictionnary that represents results.

    Parameters
    ----------
    cls : Class from wich to extract the information.

    Returns
    -------
    dict :
        Dictionnary extracted using the format defined by
        a custom sphinx directive
        with the following format::

          {attribute1: {'section': section name,
                        'units': attribute units,
                        'short': attribute short description,
                        'latexname': latex name for the attribute,
                        'symbol': symbol to represent the value}}

    """

    d = OrderedDict()
    items = [name for name in sorted(cls.__dict__.keys())]
    for attribute_name in items:
        # Extract the doc from the Processing methods
        doc = getattr(cls, attribute_name).__doc__
        if not doc:
            continue

        # All the lines in the docstring
        lines = [s.strip() for s in doc.splitlines()]

        # to store the relevant tag lines
        tag_lines = []
        n = len(lines)

        flag = False
        for line in lines:
            # Get only those that are relevant (start with .. emva1288::)
            if line.startswith('.. emva1288::'):
                flag = True
                continue
            if flag:
                if not line.strip():
                    flag = False
                    continue
                tag_lines.append(line)
        # if there are not relevant lines skip and go to next method
        if not tag_lines:
            continue

        # To store the info from the doc
        attribute_info = {}
        for line in tag_lines:
            tags = [x.strip() for x in line.split(':', 2) if x.strip()]
            # Each valid tag has to be xx:yy
            if len(tags) != 2:
                continue
            # Fill the dict
            attribute_info[tags[0].lower()] = tags[1]

        # If there is no section set it as other
        attribute_info.setdefault('section', 'other')
        d[attribute_name] = attribute_info
    return d


def _sections_first(dct):
    """For backwards compatibility where we use to have
    results with section as first keys
    """
    d = OrderedDict()

    sections = sorted({k['section'] for k in dct.values()})
    for section in sections:
        d[section] = OrderedDict()
        for k, v in dct.items():
            if v['section'] != section:
                continue
            d[section][k] = v
    return d


def obj_to_dict(obj):
    '''
    Get the info dict from the object class
    Add the values or Data to this dict

    for each method if the return value is a dict, it is inserted as
    d[SectionName][MethodName][Data] = ReturnValue
    if not
    d[SectionName][MethodName][Value] = ReturnValue
    '''
    d = cls_1288_info(obj.__class__)
    for attribute in d.keys():
        # Get the value for the given attribute
        val = getattr(obj, attribute)
        if callable(val):
            val = val()

        if isinstance(val, dict):
            d[attribute]['data'] = val
        else:
            d[attribute]['value'] = val
    return d


def dict_to_xml(d, root='results', filename=None):
    '''
    Takes a dict and return a well formed xml string
    '''

    def key_to_xml(d, r):
        '''
        Recursive call to add the key/value from dict to the r element tree
        If the value is an array joint the values casted as strings
        with whitespaces separator
        if the value is something else, it is casted as string
        '''
        for k in d.keys():
            e = etree.SubElement(r, k)
            # when the value for the key is a dict, call the function again
            if isinstance(d[k], dict):
                r.append(key_to_xml(d[k], e))

            # if the value is an array
            # add the values of the array as a string separated by whitespaces
            # Note to self: add other array types as needed
            elif isinstance(d[k], np.ndarray):
                a = [str(x) for x in d[k]]
                e.text = ' '.join(a)

            # if something else, just add the corresponding string value
            else:
                e.text = str(d[k])
        return r

    tree = etree.Element(root)
    xml = key_to_xml(d, tree)

    t = etree.tostring(xml, pretty_print=True)
    if filename is None:
        return t
    with open(filename, 'w') as f:
        f.write(t.decode('utf-8'))


def xml_to_dict(xml):
    '''
    If xml is a file, opens and parse, if string, parse it from string
    Convert the xml to a dict using element_to_dict
    Process the resulting dict:
    Cast Data to numpy float arrays (split the string by whitespaces)
    Cast Value to float
    '''
    try:
        if os.path.isfile(xml):
            tree = etree.parse(xml)
        else:
            tree = etree.ElementTree.fromstring(xml)
    except:
        print('Problems loading XML')
        return None

    def element_to_dict(r):
        '''
        Recursive call to add dictionnary elements from the r xml element
        '''
        dout = {}
        for child in r:
            # The resulting keys are forced to lowercase to compoensate
            # for some windows versions that use CamelCase
            if list(child):
                # if the element has children call the function again with
                # children as r
                dout[child.tag.lower()] = element_to_dict(child)
            else:
                dout[child.tag.lower()] = child.text
        return dout

    root = tree.getroot()
    d = element_to_dict(root)
    # loop to reconstruct arrays from strings in Data elements
    for section, method in d.items():
        for methodname, value in method.items():
            if 'data' in value:
                for data in value['data']:
                    v = value['data'][data]
                    v = v.strip()
                    v = v.split()
                    d[section][methodname]['data'][data] = np.asfarray(v)
            else:
                v = value['value']
                if v in ('None', None, 'none'):
                    v = None
                else:
                    # sometimes the decimal point is written with , instead of.
                    v = v.replace(',', '.')
                    v = float(v)
                d[section][methodname]['value'] = v
    return d


def round_significant(v, sig=SIGNIFICANT_DIGITS):
    '''
    Round up to the given significant digits, used for comparison
    '''
    if v == 0.0:
        return 0.0
    return round(v, sig - np.int(np.floor(np.log10(np.abs(v)))) - 1)


round_array = np.vectorize(round_significant)


def compare_xml(x1, x2, filename=None):
    # load the xml into dicts
    f1 = xml_to_dict(x1)
    f2 = xml_to_dict(x2)
    s = ''

    # if something is wrong abort
    if f1 is None or f2 is None:
        return s

    c1 = list(f1.keys())
    c2 = list(f2.keys())

    # loop throught the combined categories
    categories = set(c1) | set(c2)
    for category in sorted(categories):
        s += '*' * 70 + '\n'
        s += category + '\n'
        s += '*' * 70 + '\n'
        # check if missing category in one of the dicts
        if category not in c1 or category not in c2:
            t1 = category in c1
            t2 = category in c2
            s += '{0:<35}'.format('PRESENT')
            s += '{0:<20}{1:<20}FAIL'.format(str(t1), str(t2)) + '\n'
            continue

        m1 = f1[category].keys()
        m2 = f2[category].keys()
        # loop throught the combined methodnames
        methodnames = set(m1) | set(m2)
        for methodname in sorted(methodnames):
            s += '{0:<35}'.format(methodname)

            # check if methodname in dict
            if methodname not in m1:
                v1 = None
                a1 = None
            # get the value and the data
            else:
                v1 = f1[category][methodname].get('value', None)
                a1 = f1[category][methodname].get('data', None)

            if methodname not in m2:
                v2 = None
                a2 = None
            else:
                v2 = f2[category][methodname].get('value', None)
                a2 = f2[category][methodname].get('data', None)

            # first we check for values and then for data
            # if both present only values will be taken in count for comparison
            # If both are values
            if v1 is not None and v2 is not None:
                try:
                    r = (round_significant(v1) - round_significant(v2)) == 0.0
                except:
                    r = False
                t1 = v1
                t2 = v2

            # if both are arrays
            elif a1 is not None and a2 is not None:
                k1 = a1.keys()
                k2 = a2.keys()
                t1 = 'Array'
                t2 = 'Array'

                # if different keys, is invalid
                if (set(k1) ^ set(k2)):
                    r = False
                else:
                    # loop throught the keys
                    for k in k1:
                        try:
                            r = np.max(np.abs(round_array(a1[k]) -
                                              round_array(a2[k]))) == 0.0
                        except:
                            r = False

                        if not r:
                            break
            else:
                t1 = str(v1)
                t2 = str(v2)
                r = False

            s += '{0:<20}{1:<20}'.format(t1, t2)
            if r:
                s += 'OK' + '\n'
            else:
                s += 'FAIL' + '\n'

        s += '\n'

    if filename is None:
        return s
    with open(filename, 'w') as f:
        f.write(s)
        return s


def high_pass_filter(img, dim):
    """High pass filtering on image

    Computes the highpass filtering as defined by the emva standart keeping the
    result as an int image. The computation preserves the image as ints and
    doesn't perform the final division, but returns the dividing factor

    Parameters
    ----------
    img : np.array
        the image to filter
    dim : int
        The size of the highpass filter

    Returns
    -------
    d : dict
        The data dictionary of the result. The keys are:
        'img' : The filtered image
        'multiplicator' : Factor from the convolution to be considered in the
                          computation of the histogram
    """
    # By definition in the emva standart, the high pass filter is using
    # the following operation :
    #
    # y' = y-R (*) y
    #
    # where (*) represents a convolution and R is a 5x5 matrix filled with 1/25
    # In order to maintain int values for a higher precision, we make the
    # following manipulation :
    #
    # y' = P (*) y - R (*) y, where P is a 5x5 matrix of zeros with P[2,2] = 1
    #
    # multiplying the right side of the equation by 25 and then dividing by 25
    #
    # y' =  ((25 * P - R') (*) y) /25, where R' is a 5x5 matrix of ones
    #
    # simplifying
    #
    # y' = (M (*) y)/25, where M is a 5x5 matrix of -1 with M[2,2] = 24
    #
    # We don't peform the division to keep the int precision, but we return the
    # factor (25)

    # If the image is from a color camera, the format of the data to compute is
    # a masked array. Convolution cannot typically be performed on a masked
    # array, but considering this particular operation we can work around this
    # problem.
    # Considering the kernel, the unmasked values are correctly computed using
    # convolve(), while the others are wrong. But by re-applying the mask, we
    # work around this problem.
    # The multiplicator must be changed to the number of valid pixels
    # being considered in the convolution.

    # Initializing the kernel
    kernel = np.ones((dim, dim))
    if isinstance(img, ma.MaskedArray):
        # Saving the mask
        mask = ma.getmask(img)
        # These next few lines are to determine the number of pixels being
        # considered in the convolution
        value = np.ones(img.shape)
        # Applying the mask on a matrix of ones and filling the masked values
        # with 0
        value = ma.MaskedArray(value, mask)
        value = value.filled(0)
        # The convolution of the matrix with a 5x5 matrix of ones will return
        # a matrix where the values where there was no mask correspong to the
        # number of considred pixels
        value = convolve(value, kernel)
        value = ma.MaskedArray(value, mask)[2:-2, 2:-2]
        # After applying the mask all the values should be equal. If not, the
        # filter is not bayer and will not work properly with this code
        if value.max() != value.min():
            raise ValueError('Masked image does not haves a valid bayer'
                             'filter')
        # Setting the kernel
        kernel *= -1
        kernel[2, 2] = value.max()-1
        # Finally computing the filtering and slicing off the edges
        img = img.filled(0)
        img = convolve(img, kernel)
        img = np.ma.array(img, mask=mask)[2:-2, 2:-2]
    else:
        # Set the kernel and compute, then slice off the edges
        kernel *= -1
        kernel[2, 2] = dim*dim-1
        img = convolve(img, kernel)[2:-2, 2:-2]
    return {'img': img, 'multiplicator': kernel[2, 2]}



# -*- coding: utf-8 -*-
# Copyright (c) 2014 The EMVA1288 Authors. All rights reserved.
# Use of this source code is governed by a GNU GENERAL PUBLIC LICENSE that can
# be found in the LICENSE file.

"""EMVA 1288 Package initialisation file"""

from emva1288.process.data import Data1288
from emva1288.process.loader import LoadImageData
from emva1288.process.parser import ParseEmvaDescriptorFile
from emva1288.process.results import Results1288
from emva1288.process.plotting import Plotting1288


class Emva1288(object):
    def __init__(self, fname):
        parser = ParseEmvaDescriptorFile(fname)
        imgs = LoadImageData(parser.images)
        dat = Data1288(imgs.data)
        self._results = Results1288(dat.data)

    def results(self):
        self._results.print_results()

    def plot(self, *plots):
        plot = Plotting1288(self._results)
        plot.plot(*plots)

    def xml(self, filename=None):
        return self._results.xml(filename)



# This example shows how to extend/modify the plots generated

from emva1288.process import Emva1288
from emva1288.camera.dataset_generator import DatasetGenerator
from emva1288.process.plotting import Emva1288Plot, EVMA1288plots


# Create a plot
class PRNUmage(Emva1288Plot):
    name = 'PRNU Image'

    def plot(self, test):
        self.ax.imshow(test.spatial['avg'],
                       aspect='auto',
                       gid='%d:data' % test.id)


# Get a results object
dataset_generator = DatasetGenerator(width=100,
                                     height=50,
                                     bit_depth=8,
                                     dark_current_ref=30)
fname = dataset_generator.descriptor_path
e = Emva1288(fname)

######
# To plot only the PRNUImage
e.plot(PRNUmage)

#####
# To plot all plus PRNUImage
plots = EVMA1288plots + [PRNUmage]
e.plot(*plots)



import numpy as np


def qe(wavelength):
    """Simulate quantum efficiency for a specific wavelengths.

    Parameters
    ----------
    wavelength : float
                 The wavelength to compute the quantum efficency for.

    Returns
    -------
    float :
        The simulated quantum efficiency.
    """
    # For the time being we just simulate a simple gaussian
    s = 0.5
    u = 0.
    min_ = 350
    max_ = 800
    w = -1 + (wavelength) / (max_ - min_)
    qe = -.1 + np.exp((-(w - u) ** 2) /
                      (2 * s ** 2)) / (np.sqrt(np.pi * 2 * s ** 2))
    if qe < 0:
        return 0
    return qe


def nearest_value(value, array):
    """Returns the nearest value in vals.

    Parameters
    ----------
    value : float
            The value we want to get as near as possible.
    array : array_like
            The array containing the available values to get near the value.

    Returns
    -------
    The nearest element of `array` to `value`.
    """
    # http://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array
    idx = (np.abs(array - value)).argmin()
    return array[idx]


def get_irradiance(radiance, f):
    """Get The irradiance, in w/cm^2.

    Parameters
    ----------
    radiance : float
               The radiance (in W/sr/cm^2) to compute the irradiance from.
    f : float
        The f number of the setup.

    Returns
    -------
    float :
        The irradiance in W/cm^2
    """
    j = np.pi * radiance / (1 + ((2 * f) ** 2))
    return j


def get_photons(exposure, wavelength, radiance, pixel_area, f_number):
    """Get the number of photons hitting one pixel.

    Parameters
    ----------
    exposure : float
               The pixel exposure time (in ns) to the light.
    wavelength : float
                 The light wavelength hitting the pixel (in nm).
    radiance : float
               The radiance hitting the sensor (in W/sr/cm^2).
    pixel_area : float
                 The pixel area in um ^ 2
    f_number : float
               The f number of the setup.

    Returns
    -------
    float :
        The number of photons that hit the pixel.
    """
    h = 6.63e-34
    c = 3.00e8
    w = wavelength * 1e-9
    t = exposure * 1e-9
    a = pixel_area * 1e-12
    j = get_irradiance(radiance, f_number)
    return j * a * t * w / (h * c)


def get_radiance(exposure, wavelength, photons, pixel_area, f_number):
    """From the number of photons, get the radiance hitting a pixel.

    Parameters
    ----------
    exposure : float
               The pixel exposure time to the light (in ns).
    wavelength : float
                 The photons' wavelength (in nm).
    photons : float
              The number of photons that hit the pixel.
    pixel_area : float
                 The pixel area in um^2.
    f_number : float
               The f number of the setup.

    Returns
    -------
    float :
        The radiance that hit the pixel and gave the number of photons.
    """
    h = 6.63e-34
    c = 3.00e8
    w = wavelength * 1e-9
    t = exposure * 1e-9
    a = pixel_area * 1e-12

    #  p = j * a * t * w / (h * c)
    # j = np.pi * radiance / (1 + ((2 * f) ** 2))

    j = photons * h * c / (a * t * w)
    r = j * (1 + ((2 * f_number) ** 2)) / np.pi
    return r


def get_tile(arr, height, width):
    """From an array with a pattern, repeat the array to fill an 2D array
       with the given dimention.

    Parameters
    ----------
    arr : array
          The pattern wanted to be replecate in a bigger shape.
    height : int
             The height of the array wanted in return.
    width : int
            The width of the array wanted in return.

    Returns
    -------
    array :
           The pattern given(arr) replecate in the size given.
    """

    #  To reduce the execution time, we will reduce the width an height of
    #  arguments given in the np.tile function. We take the width and height
    #  wanted, dived by size of the array dimention. By careful, for an 2D
    #  array, the shape[1] is the width and the shape[0] is the height, but
    #  for an 1D array the height is 1 so the shape[0] is the width.
    #  The size need to be a int, so we use np.floor to arroud the division
    #  and add one to be sure than the size is not to short. To be sure to
    #  return the array with the good shape we use [:height, :width].

    # if the array given is an 1D array.
    if len(arr.shape) == 1:
        # the height suppose to be 1
        h_r = height
        # only the width change
        w_r = int(np.floor(width / arr.shape[0]) + 1)

    # if the array given is a 2D array.
    if len(arr.shape) == 2:
        # number of repetition needed
        h_r = int(np.floor(height / arr.shape[0]) + 1)
        w_r = int(np.floor(width / arr.shape[1]) + 1)

    # the case if the array given is an multidimentinal array.
    # else:
        # TODO: make a error message for to many dimention.

    # create a array with the dimension given and the array given
    tile = np.tile(arr, (h_r, w_r))[:height, :width]

    # if we want an 1D array so with a shape of (width,) and not (1, width)
    if height == 1:
        # tile = array([[0,0,0,...]]) so we only want the first line.
        return tile[0]
    return tile


def get_bayer_filter(t00, t01, t10, t11, width, height):
    """From different values of transmition and the size, get a bayer filter.

    +-----------+-----------+-----------+-----------+-----------+-----------+
    |              Pattern                                                  |
    +===========+===========+===========+===========+===========+===========+
    |    This case          |     Example           | Suggested values      |
    +-----------+-----------+-----------+-----------+-----------+-----------+
    |   t00     |  t01      |   G1      |   R       |    1      |  0.15     |
    +-----------+-----------+-----------+-----------+-----------+-----------+
    |   t10     |  t11      |   B       |   G2      |  0.02     |   1       |
    +-----------+-----------+-----------+-----------+-----------+-----------+

    Where *G1* and *G2* are transmition values for a green filter, *R* for a
    red filter and *B* for a blue filter. The suggested values are values
    for a standard bayer filter like the example.

    Parameters
    ----------
    t00 : float
               The transmition, in pourcentage, of the first pixel.
    t01 : float
               The transmition, in pourcentage, of the second pixel.
    t10 : float
               The transmition, in pourcentage, of the third pixel.
    t11 : float
               The transmition, in pourcentage, of the fourth pixel.
    width : int
            The number of columns in the the image.
    height : int
            The number of rows in the image.

    Returns
    -------
    array :
        The array with the bayer filter of the size gived.
    """
    pattern_rep = np.array([[t00, t01], [t10, t11]])
    b_filter = get_tile(pattern_rep, height, width)
    return b_filter



import jinja2
import os
import shutil
from distutils.dir_util import copy_tree
from collections import OrderedDict
import posixpath
from matplotlib.figure import Figure
from matplotlib.backends.backend_pdf import FigureCanvas
import numpy as np

from emva1288.process import Results1288
from emva1288.process.plotting import EVMA1288plots


def info_setup(**kwargs):
    """Container for setup information.

    All kwargs are used to update the setup information dictionary.

    Returns
    -------
    dict : A dictionary containing setup informations.
    The keys are:

      - *'Light source'* : The light source type (e.g. integrating sphere).
      - *'Light source non uniformity'* : The light source
        introducing non uniformity.
      - *'Irradiation calibration accuracy'* : The irradiation calibration
        incertainty.
      - *'Irradiation measurement error'* : The irradiation measurement
        incertainty.
      - *'Standard version'* : The EMVA1288 standard version number used.
    """
    s = OrderedDict()
    s['Light source'] = None
    s['Light source non uniformity'] = None
    s['Irradiation calibration accuracy'] = None
    s['Irradiation measurement error'] = None
    s['Standard version'] = None
    s.update(kwargs)
    return s


def info_basic(**kwargs):
    """Container for basic information.

    All kwargs are used to update the basic information dictionary for the
    report.

    Returns
    -------
    dict : A dictionary containing basic informations for the report.
    The keys are:

      - *'vendor'* : The vendor name that manufactures the camera.
      - *'model'* : The model of the tested camera.
      - *'data_type'* : The label given to the data used for the test.
      - *'sensor_type'*: The type of the tested sensor within the camera.
      - *'sensor_diagonal'* : The number of pixel in the sensor diagonal.
      - *'lens_category'* : The lens category used for the test.
      - *'resolution'* : The camera's resolution.
      - *'pixel_size'* : The sensor's pixel size.
      - *'readout_type'* : The readout type of the sensor (for CCD sensors).
      - *'transfer_type'* : The transfer type of the sensor (for CCDs).
      - *'shutter_type'* : The shutter type of the sensor (for CMOS sensors).
      - *'overlap_capabilities'* : The overlap capabilities of the sensor
        (for CMOS sensors).
      - *'maximum_readout_rate'* : The camera's maximal readout rate.
      - *'dark_current_compensation'* : If the camera support dark current
        compensation, specify it in this entry.
      - *'interface_type'* : The camera's interface type.
      - *'qe_plot'* : The sensor's quantum efficency plots.
    """
    b = {'vendor': None,
         'model': None,
         'data_type': None,
         'sensor_type': None,
         'sensor_diagonal': None,
         'lens_category': None,
         'resolution': None,
         'pixel_size': None,
         #########
         # For CCD
         'readout_type': None,
         'transfer_type': None,
         # For CMOS
         'shutter_type': None,
         'overlap_capabilities': None,
         #########
         'maximum_readout_rate': None,
         'dark_current_compensation': None,
         'interface_type': None,
         'qe_plot': None
         }
    b.update(kwargs)
    return b


def info_marketing(**kwargs):
    """Container for marketing informations.

    All kwargs are used to update the returned dictionary containing the
    marketing informations.

    Returns
    -------
    dict : A dictionary containing the marketing informations.
    The keys are:

      - *'logo'* : The path to the logo icon.
      - *'watermark'* : A text that will be printed on every page of the
        report in the background in transparent red.
      - *'missingplot'* : The path to a missing plot icon.
      - *'cover_page'* : The path to a custom cover page for the report.
    """
    m = {'logo': None,
         'watermark': None,
         'missingplot': None,
         'cover_page': None
         }
    m.update(kwargs)
    return m


def info_op():
    """Container for operation points informations.

    The returned dictionary must be filled after calling this function.

    Returns
    -------
    dict : An empty dictionary with the following keys:

      - *'name'* : The test name.
      - *'id'* : The test id.
      - *'summary_only'* : True or False, tells if, for a specific OP, the
        report should do a summary of the test instead of a full description.
      - *'results'* : The results of the test.
      - *'camera_settings'* : (OrderedDict) the dictionary of the camera's
        settings for the test.
      - *'test_parameters'* : (OrderedDict) the dictionary of the other test
        parameters.
    """
    d = {'name': None,
         'id': None,
         'summary_only': None,
         'results': None,
         'camera_settings': OrderedDict(),
         'test_parameters': OrderedDict()}
    return d


_CURRDIR = os.path.abspath(os.path.dirname(__file__))


class Report1288(object):
    """Class that has the purpose of creating a pdf report of one or more
    optical tests. This class only creates the report TeX files using the
    templates. TeX files must be compiled afterwards to generate the
    pdf files.
    """
    def __init__(self,
                 outdir,
                 setup=None,
                 basic=None,
                 marketing=None,
                 cover_page=False):
        """Report generator init method.

        Informations stored in the report can be specified by the kwargs
        passed to the object. It can be useful to use the
        :func:`info_marketing`, :func:`info_setup` and :func:`info_basic`
        functions to generate the corresponding dictionaries.
        The report generator uses `jinja2` to render the templates. Upon init,
        it calls the :meth:`template_renderer` method to get the `jinja2`
        object that will interact with the templates. Then it creates the
        output directories and files that will contain the output files.
        To create the report for different operating point/tests, one must
        call the :meth:`add` method to add each test he wants to publish in
        the report. Then, to conclude the report, he must call the
        :meth:`latex` method to generate the TeX files.

        Parameters
        ----------
        outdir : str
                 The path to the directory that will contain the report files.
        setup : dict, optional
                A dictionary containing the setup informations. If None, the
                report uses the dictionary of the :func:`info_setup`
                function.
        basic : dict, optional
                A dictionary containing basic informations about the test. If
                None, the report generator takes the dictionary from the
                :func:`info_basic` function.
        marketing : dict, optional
                    A dictionary containing
        cover_page : str, optional
                     The path to the cover page for the report. If False,
                     no cover page will be included in the report.
        """
        self._outdir = os.path.abspath(outdir)

        self.renderer = self.template_renderer()
        self.ops = []
        self.marketing = marketing or info_marketing()
        self.basic = basic or info_basic()
        self.setup = setup or info_setup()
        self.cover_page = cover_page
        self._make_dirs(outdir)

    @staticmethod
    def template_renderer(dirname=None):
        """Method that creates the renderer for the TeX report file.

        Uses the :class:`jinja2:jinja2.Environment` object to create
        the renderer. Also defines some filters for the environment
        for the missing numbers and general missings.

        Parameters
        ----------
        dirname : str, optional
                  The path to the template directory containing the TeX
                  templates. If None, it will get the templates from the
                  `./templates/` directory.

        Returns
        -------
        The renderer.
        """
        if not dirname:
            dirname = os.path.join(_CURRDIR, 'templates')
        renderer = jinja2.Environment(
            block_start_string='%{',
            block_end_string='%}',
            variable_start_string='%{{',
            variable_end_string='%}}',
            comment_start_string='%{#',
            comment_end_string='%#}',
            loader=jinja2.FileSystemLoader(dirname))

        def missingnumber(value, precision):
            # Filter for missing numbers
            if value in (None, np.nan):
                return '-'
            t = '{:.%df}' % precision
            return t.format(value)

        def missingfilter(value, default='-'):
            # General filter for missing objects that are not numbers
            if value in (None, np.nan):
                return default
            return value

        renderer.filters['missing'] = missingfilter
        renderer.filters['missingnumber'] = missingnumber
        return renderer

    def _make_dirs(self, outdir):
        """Create the directory structure for the report
        If the directory exist, raise an error
        """
        try:
            os.makedirs(self._outdir)
        except FileExistsError:  # pragma: no cover
            pass
        print('Output Dir: ', self._outdir)

        files_dir = os.path.join(self._outdir, 'files')
        try:
            os.makedirs(files_dir)
        except FileExistsError:  # pragma: no cover
            pass
        currfiles = os.path.join(_CURRDIR, 'files')
        copy_tree(currfiles, files_dir)

        upload_dir = os.path.join(self._outdir, 'upload')
        try:
            os.makedirs(upload_dir)
        except FileExistsError:  # pragma: no cover
            pass

        def uploaded_file(fname, default):
            if fname:  # pragma: no cover
                shutil.copy(os.path.abspath(fname), upload_dir)
                v = posixpath.join(
                    'upload',
                    os.path.basename(fname))
            else:
                v = posixpath.join('files', default)
            return v

        self.marketing['logo'] = uploaded_file(self.marketing['logo'],
                                               'missinglogo.pdf')

        self.marketing['missingplot'] = uploaded_file(
            self.marketing['missingplot'],
            'missingplot.pdf')

        self.basic['qe_plot'] = uploaded_file(self.basic['qe_plot'],
                                              'missingplot.pdf')

    def _write_file(self, name, content):
        # write content into a file
        fname = os.path.join(self._outdir, name)
        with open(fname, 'w') as f:
            f.write(content)
        return fname

    def _stylesheet(self):
        # generate the stylesheet content
        stylesheet = self.renderer.get_template('emvadatasheet.sty')
        return stylesheet.render(marketing=self.marketing,
                                 basic=self.basic)

    def _report(self):
        # Generate the report contents
        report = self.renderer.get_template('report.tex')
        return report.render(marketing=self.marketing,
                             basic=self.basic,
                             setup=self.setup,
                             operation_points=self.ops,
                             cover_page=self.cover_page)

    def latex(self):
        """Generate report latex files.
        """

        self._write_file('emvadatasheet.sty', self._stylesheet())
        self._write_file('report.tex', self._report())

    def _results(self, data):
        return Results1288(data)

    def _plots(self, results, id_):
        """Create the plots for the report.

        The report will include all the plots contained in the
        `~emva1288.process.plotting.EVMA1288plots` list. All plots
        will be saved in pdf format in the output directory.
        """
        names = {}
        savedir = os.path.join(self._outdir, id_)
        try:
            os.mkdir(savedir)
        except FileExistsError:  # pragma: no cover
            pass
        for plt_cls in EVMA1288plots:
            figure = Figure()
            _canvas = FigureCanvas(figure)
            plot = plt_cls(figure)
            plot.plot(results)
            plot.rearrange()
            fname = plt_cls.__name__ + '.pdf'
            figure.savefig(os.path.join(savedir, fname))
            names[plt_cls.__name__] = posixpath.join(id_, fname)
        return names

    def add(self, op, data, results=None):
        """Method that adds an operation point to the report.

        The data supplied are passed through a
        :class:`~emva1288.process.results.Results1288` object to be processed
        for the report. Also creates the plots
        which will appears in the report.

        Parameters
        ----------
        op : dict
             The dictionary containing the operation point informations.
             This dictionary must absolutely contain a 'name' key.
             See the :func:`info_op` function to get an idea to what
             keys to give.
        data : dict
               The corresponding operation point data. It must be able to
               be processed by an instance of the
               :class:`~emva1288.process.results.Results1288` class.
        """
        n = len(self.ops) + 1
        op['id'] = 'OP%d' % (n)
        if not op['name']:
            op['name'] = op['id']
        if not results:
            results = self._results(data)
        op['results'] = results.results_by_section
        results.id = n
        op['plots'] = self._plots(results, op['id'])
        self.ops.append(op)



import numpy as np


def bad_pixel(cam, num_bad_pixel, bad_value=1):
    """Generation of bad pixel in a camera.

    Parameters
    ----------
    cam: Object
         The camera object who influence the simulation of
         bad pixel in the image.
    num_bad_pixel: int
                    Number of bad pixels desired in the image.
    bad_value: float, optional
               Pourcentage, 0 to 1, of the satured e^- value for each
               bad_pixel in the the camera.
               If None, the value will be the Full well capacity.

    """
    # TODO: add error message if num_bad_pixel is higger than pixel_num
    # Set the bad value at the Full well capacity for the camera if 1.
    bad_value = bad_value * cam._u_esat
    # Count the number of pixel in the camera
    pixel_num = cam.height * cam.width
    # Select the number of bad pixel in a 1D with number
    # for they futur position.
    bad_pixel_position = np.random.choice(pixel_num, num_bad_pixel)
    # change the position in 1D to 2D.
    for i in bad_pixel_position:
        # Change the DSNU of the camera with the bad values
        # at the random position generated above.
        cam.DSNU[i // cam.width, i % cam.width] = bad_value



# -*- coding: utf-8 -*-
# Copyright (c) 2014 The EMVA1288 Authors. All rights reserved.
# Use of this source code is governed by a GNU GENERAL PUBLIC LICENSE that can
# be found in the LICENSE file.

"""Plot the results
This class takes a results.Results1288 object and produces all the plots
needed to create a reference datasheet of the EMVA1288 test

"""

from __future__ import print_function
import numpy as np

from . import routines


class Emva1288Plot(object):
    """Base class for emva plots."""
    name = ""
    """The figure's name (used as title if title is none)."""

    title = None
    """The figure's title."""

    xlabel = None
    """The x axis label."""

    ylabel = None
    """The y axis label."""

    xscale = None
    """The x axis scale."""

    yscale = None
    """The y axis scale."""

    def __init__(self, figure):
        """Base class for emva plots init function.

        The only mandatory attribute is the name, the rest are for use
        in the :meth:`setup_figure` method.

        Parameters
        ----------
        figure : The :class:`matplotlib:matplotlib.figure.Figure`
                 object to plot.
        """
        self.figure = figure
        self.setup_figure()

    def setup_figure(self):
        """Simple wrapper for one plot per figure

        Takes the name, xlabel, ylabel, xscale and yscale
        for one plot case.

        If more than one plot, just overwrite as you wish.
        """
        ax = self.figure.add_subplot(111)
        if self.title:
            ax.set_title(self.title)
        else:
            ax.set_title(self.name)
        if self.xlabel:
            ax.set_xlabel(self.xlabel)
        if self.ylabel:
            ax.set_ylabel(self.ylabel)
        if self.xscale:
            ax.set_xscale(self.xscale)
        if self.yscale:
            ax.set_yscale(self.yscale)

        self.ax = ax

    def plot(self, test):
        """Method to show the figures.

        Parameters
        ----------
        test : Do nothing for this method but can be
               used for the subclass method.

        Raises
        ------
        NotImplementedError
            If this method is not overridden.

        Notes
        -----
        Must be overridden in subclasses.
        """
        raise NotImplementedError

    def set_legend(self, ax):
        """Shortcut to add legend.

        Parameters
        ----------
        ax : The :class:`matplotlib:matplotlib.axes.Axes` object to which
             the legend will be added.
        """
        ax.legend(loc='best')
        legend = ax.get_legend()
        if legend is not None:
            if getattr(legend, 'draggable', False):
                legend.draggable(True)

    def rearrange(self):
        """Opportunity to change axis or limits after all the tests have
        been plotted.

        Uses :meth:`matplotlib:matplotlib.figure.Figure.tight_layout` method.
        """
        self.figure.tight_layout()

    def reduce_ticks(self, ax, axis, n=4):
        """Reduce the number of ticks in ax.axis

        Uses the :meth:`matplotlib:matplotlib.axes.Axes.locator_params` method.

        Parameters
        ----------
        ax : The :class:`matplotlib:matplotlib.axes.Axes` object to which
             the number of ticks will be changed.
        axis : str, {'x', 'y', 'both'}
               Axis on which to operate.
        n : int, optional
            Number of bins between ticks to be left.
        """
        ax.locator_params(axis=axis, nbins=n)


class PlotSensitivity(Emva1288Plot):
    name = 'Sensitivity'
    xlabel = '$\mu_p$ [mean number of photons/pixel]'
    ylabel = '$\mu_y - \mu_{y.dark}$ [DN]'

    def plot(self, test):
        ax = self.ax
        ax.plot(test.temporal['u_p'],
                test.temporal['u_y'] - test.temporal['u_ydark'],
                label='Data',
                gid='%d:data' % test.id)

        ax.plot(test.temporal['u_p'],
                test.R * test.temporal['u_p'], '--',
                label='Fit',
                gid='%d:fit' % test.id)

        xi = test.temporal['u_p'][test.index_sensitivity_min]
        xf = test.temporal['u_p'][test.index_sensitivity_max]
        yi = (test.temporal['u_y'] -
              test.temporal['u_ydark'])[test.index_sensitivity_min]
        yf = (test.temporal['u_y'] -
              test.temporal['u_ydark'])[test.index_sensitivity_max]

        ax.plot((xi, xf), (yi, yf),
                label='Fit range',
                linestyle='None',
                marker='o',
                gid='%d:marker' % test.id)

        self.set_legend(ax)

    def rearrange(self):
        self.ax.ticklabel_format(axis='x', style='sci', scilimits=(1, 4))
        self.figure.tight_layout()


class PlotUyDark(Emva1288Plot):
    '''Create $\mu_{y.dark}$ plot'''

    name = 'Mean gray value in dark'
    title = '$\mu_{y.dark}$'
    xlabel = 'exposure time [ns]'
    ylabel = '$\mu_{y.dark}$ [DN]'

    def plot(self, test):
        ax = self.ax

        if np.mean(test.temporal['texp']) == test.temporal['texp'][0]:
            ax.plot(test.temporal['texp'],
                    test.temporal['u_ydark'],
                    'o',
                    markersize=5,
                    label='Data',
                    gid='%d:data' % test.id)
        else:
            ax.plot(test.temporal['texp'],
                    test.temporal['u_ydark'],
                    label='Data',
                    gid='%d:data' % test.id)
        ax.ticklabel_format(axis='x', style='sci', scilimits=(1, 4))
        self.set_legend(ax)


class PlotPTC(Emva1288Plot):
    '''Create Photon Transfer plot'''

    name = 'Photon Transfer'
    xlabel = '$\mu_y - \mu_{y.dark}$ [DN]'
    ylabel = '$\sigma^2_y - \sigma^2_{y.dark}$ [DN$^2$]'

    def plot(self, test):
        ax = self.ax

        X = test.temporal['u_y'] - test.temporal['u_ydark']
        Y = test.temporal['s2_y'] - test.temporal['s2_ydark']
        ax.plot(X, Y,
                label='Data',
                gid='%d:data' % test.id)

        ax.plot(X, test.K * X,
                linestyle='--',
                label='Fit',
                gid='%d:fit' % test.id)

        ax.plot((X[test.index_u_ysat], ), (Y[test.index_u_ysat], ),
                marker='s',
                linestyle='None',
                label='Saturation',
                gid='%d:marker' % test.id)

        ax.plot((X[test.index_sensitivity_min],
                 X[test.index_sensitivity_max]),
                (Y[test.index_sensitivity_min],
                 Y[test.index_sensitivity_max]),
                linestyle='None',
                marker='o',
                label='Fit range',
                gid='%d:marker' % test.id)

        self.set_legend(ax)


class PlotSNR(Emva1288Plot):
    '''Create SNR plot '''

    name = 'Signal to Noise Ratio'
    xlabel = '$\mu_{p}$ [mean number of photons/pixel]'
    ylabel = 'SNR'

    def setup_figure(self):
        super(PlotSNR, self).setup_figure()
        self.ax.loglog()
        max_ideal = []
        self.max_ideal = max_ideal

    def plot(self, test):
        ax = self.ax

        X = np.arange(test.u_p_min, test.u_p_sat,
                      (test.u_p_sat - test.u_p_min) / 100.0)

        # remove the zeros on the denominator, at saturation the temporal
        # noise is zero!
        nz = np.nonzero(test.temporal['s2_y'])
        ax.plot(test.temporal['u_p'][nz],
                (test.temporal['u_y'] - test.temporal['u_ydark'])[nz] /
                np.sqrt(test.temporal['s2_y'][nz]),
                marker='.',
                ls=' ',
                label='Data',
                gid='%d:data' % test.id)

        ax.plot(X,
                ((test.QE / 100) * X) /
                np.sqrt((test.sigma_d) ** 2 +
                        (test.s2q / (test.K) ** 2) +
                        ((test.QE / 100) * X)),
                linestyle=':',
                label='Theoretical',
                gid='%d:fit' % test.id)

        ideal = np.sqrt(X)
        self.max_ideal.append(ideal[-1])
        ax.plot((X),
                ideal,
                linestyle='-.',
                label='Ideal',
                gid='%d:fit' % test.id)

        ax.axvline(test.u_p_min,
                   label='$\mu_{p.min} = %.1f[p]$' % test.u_p_min,
                   gid='%d:marker' % test.id)

        ax.axvline(test.u_p_sat,
                   label='$\mu_{p.sat} = %.1f[p]$' % test.u_p_sat,
                   gid='%d:marker' % test.id)

        ax.plot(X,
                ((test.QE / 100) * X) /
                np.sqrt((test.sigma_d) ** 2 +
                        (test.s2q / (test.K) ** 2) +
                        ((test.QE / 100) * X) +
                        (test.DSNU1288 ** 2) +
                        (((test.PRNU1288 / 100) *
                          (test.QE / 100.) * X) ** 2)),
                linestyle='--',
                label='Total SNR',
                gid='%d:fit' % test.id)

        ax.set_ylim(1, max(self.max_ideal))
        self.set_legend(ax)


class PlotLinearity(Emva1288Plot):
    '''Create Linearity plot'''

    name = 'Linearity'
    xlabel = '$\mu_{p}$ [mean number of photons/pixel]'
    ylabel = '$\mu_y - \mu_{y.dark}$ [DN]'

    def plot(self, test):
        ax = self.ax

        X = test.temporal['u_p']
        Y = test.temporal['u_y'] - test.temporal['u_ydark']
        ax.plot(X, Y,
                label='Data',
                gid='%d:data' % test.id)

        ax.plot(X,
                test.linearity()['fit_slope'] *
                X + test.linearity()['fit_offset'],
                linestyle='--',
                label='Fit',
                gid='%d:fit' % test.id)

        ax.plot((X[test.index_linearity_min], X[test.index_linearity_max]),
                (Y[test.index_linearity_min], Y[test.index_linearity_max]),
                label='Fit range',
                linestyle='None',
                marker='o',
                gid='%d:marker' % test.id)

        self.set_legend(ax)
        self.ax.ticklabel_format(axis='x', style='sci', scilimits=(1, 4))


class PlotDeviationLinearity(Emva1288Plot):
    '''Create Deviation Linearity plot'''

    name = 'Deviation linearity'
    xlabel = '$\mu_{p}$ [mean number of photons/pixel]'
    ylabel = 'Linearity error LE [%]'

    def plot(self, test):
        ax = self.ax

        X = test.temporal['u_p'][test.index_linearity_min:
                                 test.index_linearity_max]
        deviation = test.linearity()['relative_deviation']
        Y = deviation[test.index_linearity_min: test.index_linearity_max]
        ax.plot(X, Y,
                label='Data',
                gid='%d:data' % test.id)

        ax.plot((X[0], X[-1]),
                (Y[0], Y[-1]),
                label='Fit range',
                linestyle='None',
                marker='o',
                gid='%d:marker' % test.id)

        self.set_legend(ax)
        self.ax.ticklabel_format(axis='x', style='sci', scilimits=(1, 4))


class PlotHorizontalSpectrogramPRNU(Emva1288Plot):
    '''Create Horizontal spectrogram PRNU plot'''

    name = 'Horizontal spectrogram PRNU'
    xlabel = 'cycles [periods/pixel]'
    ylabel = 'Standard deviation and\nrelative presence of each cycle [%]'
    yscale = 'log'

    def plot(self, test):
        ax = self.ax

        data = (test.spatial['sum'] - test.spatial['sum_dark'])
        data = data / test.spatial['L']
        data_mean = np.mean(data)

        spectrogram = routines.FFT1288(data) / data_mean

        ax.plot(routines.GetFrecs(spectrogram),
                (np.sqrt(spectrogram)),
                label='Data',
                gid='%d:data' % test.id)

        ax.axhline(test.PRNU1288,
                   label='$PRNU_{1288}$',
                   linestyle='--',
                   color='r',
                   gid='%d:marker' % test.id)

        ax.axhline(100 * np.sqrt(test.sigma_2_y_stack) / data_mean,
                   label='$\sigma^2_{y.stack}$',
                   linestyle='-.',
                   color='g',
                   gid='%d:marker' % test.id)

        self.set_legend(ax)


class PlotHorizontalSpectrogramDSNU(Emva1288Plot):
    '''Create Horizontal spectrogram DSNU plot'''

    name = 'Horizontal spectrogram DSNU'
    xlabel = 'cycles [periods/pixel]'
    ylabel = 'Standard deviation and\nrelative presence of each cycle [DN]'
    yscale = 'log'

    def plot(self, test):
        ax = self.ax

        spectrogram = routines.FFT1288(test.spatial['sum_dark'],
                                       n=test.spatial['L'])
        ax.plot(routines.GetFrecs(spectrogram),
                np.sqrt(spectrogram),
                label='Data',
                gid='%d:data' % test.id)

        ax.axhline(test.DSNU1288_DN(),
                   label='$DSNU_{1288.DN}$',
                   linestyle='--',
                   color='r',
                   gid='%d:marker' % test.id)

        ax.axhline(np.sqrt(test.sigma_2_y_stack_dark),
                   label='$\sigma^2_{y.stack.dark}$',
                   linestyle='-.',
                   color='g',
                   gid='%d:marker' % test.id)

        self.set_legend(ax)


class PlotVerticalSpectrogramPRNU(Emva1288Plot):
    '''Create Vertical spectrogram PRNU plot'''

    name = 'Vertical spectrogram PRNU'
    xlabel = 'cycles [periods/pixel]'
    ylabel = 'Standard deviation and\nrelative presence of each cycle [%]'
    yscale = 'log'

    def plot(self, test):
        ax = self.ax
        data = test.spatial['sum'] - test.spatial['sum_dark']
        data = data / test.spatial['L']
        data_mean = np.mean(data)
        spectrogram = routines.FFT1288(data, rotate=True) / data_mean

        ax.plot((routines.GetFrecs(spectrogram)),
                (np.sqrt(spectrogram)),
                label='Data',
                gid='%d:data' % test.id)

        ax.axhline(test.PRNU1288,
                   label='$PRNU_{1288}$',
                   linestyle='--',
                   color='r',
                   gid='%d:marker' % test.id)

        ax.axhline(100 * np.sqrt(test.sigma_2_y_stack) / data_mean,
                   label='$\sigma^2_{y.stack}$',
                   linestyle='-.',
                   color='g',
                   gid='%d:marker' % test.id)

        self.set_legend(ax)


class PlotVerticalSpectrogramDSNU(Emva1288Plot):
    '''Create Vertical spectrogram DSNU plot'''

    name = 'Vertical spectrogram DSNU'
    xlabel = 'Cycles [periods/pixel]'
    ylabel = 'Standard deviation and\nrelative presence of each cycle [DN]'
    yscale = 'log'

    def plot(self, test):
        ax = self.ax

        spectrogram = routines.FFT1288(test.spatial['sum_dark'],
                                       rotate=True,
                                       n=test.spatial['L'])
        ax.plot(routines.GetFrecs(spectrogram),
                np.sqrt(spectrogram),
                label='Data',
                gid='%d:data' % test.id)

        ax.axhline(test.DSNU1288_DN(),
                   label='$DSNU_{1288.DN}$',
                   linestyle='--',
                   color='r',
                   gid='%d:marker' % test.id)

        ax.axhline(np.sqrt(test.sigma_2_y_stack_dark),
                   label='$\sigma^2_{y.stack.dark}$',
                   linestyle='-.',
                   color='g',
                   gid='%d:marker' % test.id)

        self.set_legend(ax)


class PlotLogarithmicHistogramDSNU(Emva1288Plot):
    '''Create Logarithmic histogram DSNU plot'''

    name = 'Logarithmic histogram DSNU'
    xlabel = 'Dark value [DN]'
    ylabel = 'Number of pixels'
    yscale = 'log'

    def plot(self, test):
        ax = self.ax

        hist = test.histogram_DSNU

        ax.plot(hist['bins'], hist['values'],
                gid='%d:data' % test.id,
                label='Data')
        ax.plot(hist['bins'], hist['model'],
                '--',
                gid='%d:fit' % test.id,
                label='Model')

        self.set_legend(ax)

        ax.axis(ymin=1.0, ymax=np.max(hist['values']) * 2)


class PlotLogarithmicHistogramPRNU(Emva1288Plot):
    '''Create Logarithmic histogram PRNU plot'''

    name = 'Logarithmic histogram PRNU'
    xlabel = 'Deviation from the mean [DN]'
    ylabel = 'Number of pixels'
    yscale = 'log'

    def plot(self, test):
        ax = self.ax
        hist = test.histogram_PRNU

        ax.plot(hist['bins'], hist['values'],
                gid='%d:data' % test.id,
                label='Data')
        ax.plot(hist['bins'], hist['model'], '--',
                gid='%d:fit' % test.id,
                label='Model')

        self.set_legend(ax)
        ax.axis(ymin=0.5, ymax=np.max(hist['values']) * 2)


class PlotAccumulatedLogHistogramDSNU(Emva1288Plot):
    '''Create Accumulated log histogram DSNU plot'''

    name = 'Accumulated log histogram DSNU'
    xlabel = 'Minimal deviation from the mean [DN]'
    ylabel = 'Number of pixels'
    yscale = 'log'

    def plot(self, test):
        ax = self.ax
        hist = test.histogram_DSNU_accumulated

        ax.plot(hist['bins'], hist['values'],
                gid='%d:data' % test.id,
                label='Data')

        self.set_legend(ax)
        self.figure.tight_layout()


class PlotAccumulatedLogHistogramPRNU(Emva1288Plot):
    '''Create Accumulated log histogram PRNU plot'''

    name = 'Accumulated log histogram PRNU'
    xlabel = 'Minimal deviation from the mean [%]'
    ylabel = 'Number of pixels'
    yscale = 'log'

    def plot(self, test):
        ax = self.ax

        hist = test.histogram_PRNU_accumulated

        ax.plot(hist['bins'], hist['values'],
                gid='%d:data' % test.id,
                label='Data')

        self.set_legend(ax)


class ProfileBase(Emva1288Plot):
    def __init__(self, *args, **kwargs):
        Emva1288Plot.__init__(self, *args, **kwargs)
        # Dict to keep track of max and min values to adjust the
        # limit of the axis
        self.axis_limits = {'bright': {'max': [], 'min': [], 'length': []},
                            'dark': {'max': [], 'min': [], 'length': []}}

    def _get_extremes(self, mean_, min_, max_):
        min_min_i = np.argmax(mean_ - min_)
        min_min = min_[min_min_i]
        min_perc = np.abs(100. - (min_min * 100. / mean_[min_min_i]))
        min_label = 'Min ({:.1f} {:.1f}%)'.format(min_min,
                                                  min_perc)

        max_max_i = np.argmax(max_ - min_)
        max_max = max_[max_max_i]
        max_perc = np.abs(100. - (max_max * 100. / mean_[max_max_i]))
        max_label = 'Max ({:.1f} {:.1f}%)'.format(max_max,
                                                  max_perc)

        return {'min_deviation': min_min, 'min_precentage': min_perc,
                'min_label': min_label, 'max_label': max_label,
                'max_deviation': max_max, 'max_percentage': max_perc}

    def _get_image_profiles(self, image):
        """Get the images profiles. Supports the masked arrays.

        Masked arrays cannot be shown directly with continuous lines.
        To fix this, we only return the profiles excluding the masks
        constants.

        Also returns the x arrays corresponding the the profile.
        """
        img = image
        if self.vertical:
            img = np.transpose(image)

        profile = np.mean(img, axis=0)
        profile_min = np.min(img, axis=0)
        profile_max = np.max(img, axis=0)

        mid_i = np.shape(img)[0]
        profile_mid = img[mid_i // 2, :]

        # Verifications for profile_mid
        if isinstance(profile_mid, np.ma.core.MaskedArray):
            if profile_mid.mask.all():
                # If by chance mid is a column of masked constant,
                # take the next one
                profile_mid = img[mid_i // 2 + 1, :]

        # use _get_x_y to get the x and profile of unmasked values
        d = {'mean': self._get_x_y(profile),
             'min': self._get_x_y(profile_min),
             'max': self._get_x_y(profile_max),
             'mid': self._get_x_y(profile_mid),
             }
        return d

    def _get_x_y(self, profile):
        x = np.arange(len(profile))
        if isinstance(profile, np.ma.MaskedArray):
            masks = profile.mask
            # Index lists of where there is masked constants
            index = [i for i in range(len(masks)) if not masks[i]]
            # chop out masked values
            x = x[index]
            profile = profile[index]
        return (x, profile)

    def get_profiles(self, bright, dark):
        b_p = self._get_image_profiles(bright)
        # index 1 is the profile (0 is x-axis)
        b_mean = np.mean(b_p['mean'][1])
        self.axis_limits['bright']['min'].append(0.9 * b_mean)
        self.axis_limits['bright']['max'].append(1.1 * b_mean)

        d_p = self._get_image_profiles(dark)
        self.axis_limits['dark']['min'].append(0.9 * np.mean(d_p['min'][1]))
        self.axis_limits['dark']['max'].append(1.1 * np.mean(d_p['max'][1]))
        return {'bright': b_p, 'dark': d_p}

    def plot(self, test):
        # index of profiles of what will be the x and y axis
        x = 0
        y = 1
        legend_loc = 'upper right'
        if self.vertical:
            x = 1
            y = 0
            legend_loc = (0.8, 0.65)

        ax = self.ax
        ax2 = self.ax2

        L = test.spatial['L']
        bimg = (test.spatial['sum'] - test.spatial['sum_dark']) / L
        dimg = test.spatial['sum_dark'] / L
        profiles = self.get_profiles(bimg, dimg)

        # to keep the lines number for legend
        bright_plots = []
        labels = []

        for typ, color in (('mid', 'green'),
                           ('min', 'blue'),
                           ('max', 'orange'),
                           ('mean', 'black')):
            # label has first letter capital
            label = typ.capitalize()
            labels.append(label)

            # bright plot
            line = ax.plot(profiles['bright'][typ][x],
                           profiles['bright'][typ][y],
                           label=label,
                           color=color,
                           gid='%d:marker' % test.id)[0]
            bright_plots.append(line)

            # dark plot
            ax2.plot(profiles['dark'][typ][x],
                     profiles['dark'][typ][y],
                     label=label,
                     color=color,
                     gid='%d:data' % test.id)

        # Place legend
        self.figure.legend(bright_plots,
                           labels,
                           loc=legend_loc)


class PlotHorizontalProfile(ProfileBase):
    '''Create Horizontal profile plot
    Profile is done with spatial images.
    '''

    name = 'Horizontal profile'
    vertical = False

    def setup_figure(self):
        self.ax = self.figure.add_subplot(211)
        self.ax2 = self.figure.add_subplot(212)
        self.figure.suptitle(self.name)
        self.ax.set_title('PRNU')
        self.ax.set_ylabel('Vertical line [DN]')
        self.ax2.set_title('DSNU')
        self.ax2.set_xlabel('Index of the line')
        self.ax2.set_ylabel('Vertical line [DN]')

    def rearrange(self):
        self.ax.set_xticks([])
        self.ax.axis(ymin=min(self.axis_limits['bright']['min']),
                     ymax=max(self.axis_limits['bright']['max']))
        self.ax2.axis(ymin=min(self.axis_limits['dark']['min']),
                      ymax=max(self.axis_limits['dark']['max']))

        self.reduce_ticks(self.ax2, 'y')
        self.reduce_ticks(self.ax, 'y')
        self.figure.tight_layout(pad=2)


class PlotVerticalProfile(ProfileBase):
    '''Create Vertical profile plot.
    Profile is done with spatial images.
    '''

    name = 'Vertical profile'
    vertical = True

    def setup_figure(self):
        self.ax2 = self.figure.add_subplot(121)
        self.ax = self.figure.add_subplot(122)
        self.figure.suptitle(self.name)
        self.ax2.set_title('DSNU')
        self.ax2.set_xlabel('Vertical line [DN]')
        self.ax2.set_ylabel('Index of the line')
        self.ax.set_title('PRNU')
        self.ax.set_xlabel('Vertical line [DN]')

    def rearrange(self):
        self.ax.set_yticks([])
        self.ax.axis(xmin=min(self.axis_limits['bright']['min']),
                     xmax=max(self.axis_limits['bright']['max']))
        self.ax2.axis(xmin=min(self.axis_limits['dark']['min']),
                      xmax=max(self.axis_limits['dark']['max']))
        self.ax2.invert_yaxis()
        self.ax.invert_yaxis()
        self.reduce_ticks(self.ax2, 'x')
        self.reduce_ticks(self.ax, 'x')
        self.figure.tight_layout()


EVMA1288plots = [PlotPTC,
                 PlotSNR,
                 PlotSensitivity,
                 PlotUyDark,
                 PlotLinearity,
                 PlotDeviationLinearity,
                 PlotHorizontalSpectrogramPRNU,
                 PlotHorizontalSpectrogramDSNU,
                 PlotVerticalSpectrogramPRNU,
                 PlotVerticalSpectrogramDSNU,
                 PlotLogarithmicHistogramDSNU,
                 PlotLogarithmicHistogramPRNU,
                 PlotAccumulatedLogHistogramDSNU,
                 PlotAccumulatedLogHistogramPRNU,
                 PlotHorizontalProfile,
                 PlotVerticalProfile]
"""
    This list is quite exhaustive. There are the links
    to corresponding documentation:

    - :class:`~emva1288.process.plotting.PlotPTC`
    - :class:`~emva1288.process.plotting.PlotSNR`
    - :class:`~emva1288.process.plotting.PlotSensitivity`
    - :class:`~emva1288.process.plotting.PlotUyDark`
    - :class:`~emva1288.process.plotting.PlotLinearity`
    - :class:`~emva1288.process.plotting.PlotDeviationLinearity`
    - :class:`~emva1288.process.plotting.PlotHorizontalSpectrogramPRNU`
    - :class:`~emva1288.process.plotting.PlotHorizontalSpectrogramDSNU`
    - :class:`~emva1288.process.plotting.PlotVerticalSpectrogramPRNU`
    - :class:`~emva1288.process.plotting.PlotVerticalSpectrogramDSNU`
    - :class:`~emva1288.process.plotting.PlotLogarithmicHistogramDSNU`
    - :class:`~emva1288.process.plotting.PlotLogarithmicHistogramPRNU`
    - :class:`~emva1288.process.plotting.PlotAccumulatedLogHistogramPRNU`
    - :class:`~emva1288.process.plotting.PlotAccumulatedLogHistogramDSNU`
    - :class:`~emva1288:emva1288.process.plotting.PlotHorizontalProfile`
    - :class:`~emva1288.process.plotting.PlotVerticalProfile`
"""


class Plotting1288(object):
    """EMVA1288 plots

    Creates and shows all plots necessary to prepare a camera or sensor
    descriptive report compliant with EMVA Standard 1288 version 3.1.
    """

    def __init__(self, *tests):
        '''
        Parameters
        ----------
        tests: list
            List of tests to Plot
        '''

        self.tests = []
        for test in tests:
            if not getattr(test, 'id', False):
                test.id = id(test)
        self.tests.append(test)

    def plot(self, *plots):
        """Plot EMVA1288 plots

        Parameters
        ----------
        plots: list
            List of plots to plot

        """
        import matplotlib.pyplot as plt
        if not plots:
            plots = EVMA1288plots

        for i, plot_cls in enumerate(plots):
            figure = plt.figure(i)
            plot = plot_cls(figure)
            for test in self.tests:
                plot.plot(test)
            plot.rearrange()
            figure.canvas.set_window_title(plot.name)
        plt.show()



import unittest
import tempfile
import os
from emva1288.process import ParseEmvaDescriptorFile, LoadImageData, Data1288
from emva1288.camera.dataset_generator import DatasetGenerator
from emva1288.report import info_op, Report1288
from emva1288.process.plotting import EVMA1288plots


class TestReportGenerator(unittest.TestCase):
    _height = 50
    _width = 100
    _bit_depth = 8
    _L = 50
    _steps = 10
    _radiance_min = None
    _exposure_max = 50000000

    def setUp(self):
        # create dataset
        self.dataset = DatasetGenerator(height=self._height,
                                        width=self._width,
                                        bit_depth=self._bit_depth,
                                        L=self._L,
                                        steps=self._steps,
                                        radiance_min=self._radiance_min,
                                        exposure_max=self._exposure_max)

        # parse dataset
        self.parser = ParseEmvaDescriptorFile(self.dataset.descriptor_path)
        # load images
        self.loader = LoadImageData(self.parser.images)
        # create data
        self.data = Data1288(self.loader.data)
        # create operation point dict
        self.op = info_op()

    def tearDown(self):
        del self.dataset
        del self.parser
        del self.loader
        del self.op
        del self.data

    def test_report_generation(self):
        with tempfile.TemporaryDirectory() as outdir:
            # create report
            report = Report1288(outdir)
            report.add(self.op, self.data.data)
            # check that output directory has been created
            for directory in ('files', 'OP1', 'upload'):
                self.assertTrue(os.path.isdir(os.path.join(outdir, directory)))
            # check that plots have been created and saved
            for plt in EVMA1288plots:
                name = plt.__name__
                path = os.path.join(outdir, 'OP1', name + '.pdf')
                self.assertTrue(os.path.isfile(path))
            # generate tex files
            report.latex()
            # check that tex files have been created
            for fil in ('emvadatasheet.sty', 'report.tex'):
                path = os.path.join(outdir, fil)
                self.assertTrue(os.path.isfile(path))



# -*- coding: utf-8 -*-
# Copyright (c) 2014 The EMVA1288 Authors. All rights reserved.
# Use of this source code is governed by a GNU GENERAL PUBLIC LICENSE that can
# be found in the LICENSE file.

"""Load image data
This class takes a dictionary (product of parser.ParseEmvaDescriptorFile).
Load the related images and reduce it's data to the minimum possible,
preserving all relevant image data in as integers
"""

import os
import numpy as np
from . import routines
import logging


class LoadImageData(object):
    """Using an info dictionary, load the images and fill data with
    the relevant information.
    """

    def __init__(self,
                 images,
                 path='',
                 loglevel=logging.INFO,
                 fload=routines.load_image,
                 fload_args=[],
                 fload_kwargs={}):
        """Loader init method.

        Uses a :class:`python:logging.Logger` object to print infos for the
        user. In the initialization, it :meth:`loads <_load_data>` the infos
        from the images into the data dictionary.

        Parameters
        ----------
        images : dict
                 The
                 :attr:`~emva1288.process.parser.ParseEmvaDescriptorFile.images`
                 attribute of a
                 :class:`~emva1288.process.parser.ParseEmvaDescriptorFile`
                 object that have parsed the images of a descriptor file.
        path : str, optional
               The path to the directory containing the images.
        loglevel : int, optional
                   The information level for the
                   :class:`~python:logging.Logger`.
        fload : func. optional
                The function that will load the images (one at a time).
        fload_args : list, optional
                     The list of args for the fload function.
        fload_kwargs : dict, optional
                       The kwargs dictionary for the fload function.
        """
        self.data = {'temporal': {}, 'spatial': {}}
        self._fload = fload
        self._fload_args = fload_args
        self._fload_kwargs = fload_kwargs
        self._shape = set()

        logging.basicConfig()
        self.log = logging.getLogger('Loader')
        self.log.setLevel(loglevel)

        self._path = path
        # load the images
        self._load_data(images)

    def _load_data(self, images):
        """Using the information in images dict fill self.data when loading
        the images and fills the temporal and spatial dicts
        """
        # iterate over the kind
        for kind in ('temporal', 'spatial'):
            exposures = set(images[kind].keys())

            # iterate over the exposure times for each kind
            for exposure in exposures:
                # for this exposure time, get the list of photons counts
                photon_counts = list(images[kind][exposure].keys())

                if 0.0 not in photon_counts:
                    # Every exposure time should have at least one dark
                    # image with a 0 photon count.
                    raise ValueError('Every exposure must have a 0.0 photons'
                                     ' for dark information')
                if len(photon_counts) < 2:
                    # Every exposure time should have at least one dark image
                    # and one bright image.
                    raise ValueError('All exposure must have at least 2 points'
                                     ' one dark and one bright')

                # Iterate over the images for this exposure time and this
                # photon count
                for photons, fnames in images[kind][exposure].items():
                    # get the image data
                    data_imgs = self._get_imgs_data(fnames, kind)
                    # fill data dict
                    self.data[kind].setdefault(exposure, {})
                    self.data[kind][exposure][photons] = data_imgs

        shape = self._shape.pop()
        self.data['height'] = shape[0]
        self.data['width'] = shape[1]

    def _get_imgs_data(self, fnames, kind):
        """Returns the desired image data
        This depends on the kind of data
        """
        # Load a list containing the images
        arr_imgs = self._load_imgs(fnames)
        # Get the sum and the pseudo-variance of this list of images
        imgs = routines.get_int_imgs(arr_imgs)
        # For spatial we only want the sum and pseudo-var of the images
        if kind == 'spatial':
            return imgs
        # For temporal, we only want the basic statistics (we don't need
        # the whole sum and pvar images)
        d = {}
        d['sum'] = np.sum(imgs['sum'])  # sum of the sum of images
        d['pvar'] = np.sum(imgs['pvar'])  # sum of the pvar image
        return d

    def _load_imgs(self, fnames):
        """For a list of path to images, load and append them to a returned list.
        """
        imgs = []
        # Iterate over all images
        for fname in fnames:
            # Get the path to the specific image.
            filename = os.path.join(self._path, fname)
            self.log.debug('Loading ' + fname)
            if not os.path.isfile(filename):
                # If the path is not good, raise an error.
                raise IOError('Not such file: ' + filename)

            # If path is good, load the image using the fload function
            img = self._fload(filename,
                              *self._fload_args,
                              **self._fload_kwargs)
            # Append the loaded image to the list
            imgs.append(img)

            # Add the image shape to the set
            self._shape.add(img.shape)
            if len(self._shape) > 1:  # pragma: no cover
                # If the shape set contains more than 1 element, this means
                # not all images have the same shape.
                raise ValueError('All images must have the same shape')

        return imgs



import unittest
from emva1288.camera.dataset_generator import DatasetGenerator
from emva1288.process.parser import ParseEmvaDescriptorFile
from emva1288.process.loader import LoadImageData
from emva1288.process.data import Data1288
from emva1288.unittests.test_routines import del_obj


class TestData(unittest.TestCase):
    _height = 50
    _width = 100
    _bit_depth = 8
    _L = 50
    _steps = 10
    _radiance_min = None
    _exposure_max = 50000000

    def _init(self):
        # create dataset
        dataset = DatasetGenerator(height=self._height,
                                   width=self._width,
                                   bit_depth=self._bit_depth,
                                   L=self._L,
                                   steps=self._steps,
                                   radiance_min=self._radiance_min,
                                   exposure_max=self._exposure_max)
        # parse dataset
        parser = ParseEmvaDescriptorFile(dataset.descriptor_path)
        # load images
        loader = LoadImageData(parser.images)
        # create data
        data = Data1288(loader.data)
        return dataset, parser, loader, data

    def test_data(self):
        """Test that data1288 retrieves information."""
        ds, p, l, d = self._init()
        self.dataset = ds
        self.parser = p
        self.loader = l
        self.data = d

        # test number of pixels
        self.assertEqual(self.data.pixels, self._height * self._width)

        # test data attribute
        #####################
        data = self.data.data

        # Test spatial
        self.assertEqual(data['spatial']['L'], self._L)
        self.assertEqual(data['spatial']['L_dark'], self._L)  # same L for dark
        # spatial exposure time
        texp = list(self.dataset.points['spatial'].keys())[0]
        self.assertEqual(data['spatial']['texp'], texp)
        # spatial photons
        radiance = self.dataset.points['spatial'][texp][0]
        photons = round(self.dataset.cam.get_photons(radiance), 3)
        self.assertEqual(data['spatial']['u_p'], photons)
        # spatial data are images
        for typ in ('sum_dark', 'sum'):
            self.assertTrue(typ in data['spatial'].keys())
            self.assertEqual(data['spatial'][typ].shape, (self._height,
                                                          self._width))

        # test temporal
        # all temporal data are arrays of length steps
        for typ in ('s2_y', 's2_ydark', 'texp', 'u_p', 'u_y', 'u_ydark'):
            self.assertTrue(typ in data['temporal'].keys())
            self.assertEqual(len(data['temporal'][typ]), self._steps)

        # test exposure times and photons have well be retrieved
        times = list(self.dataset.points['temporal'].keys())
        for i, (exp, photons) in enumerate(zip(data['temporal']['texp'],
                                               data['temporal']['u_p'])):
            time = times[i]
            radiance = self.dataset.points['temporal'][time][0]
            photon = round(self.dataset.cam.get_photons(radiance, time), 3)
            self.assertEqual(exp, times[i])
            self.assertEqual(photons, photon)

        # delete objects
        del_obj(self.dataset, self.parser, self.loader, self.data)

    def test_1exposure(self):
        """Test that when there is only one exposure time, the temporal data
        dictionary has same length than the number of photons."""
        self._radiance_min = 0.1
        self._exposure_max = 1000000
        ds, p, l, d = self._init()
        self.dataset = ds
        self.parser = p
        self.loader = l
        self.data = d
        temporal = self.data.data['temporal']
        l = len(temporal['u_p'])
        # test that all temporal data arrays have same length
        self.assertEqual(len(temporal['texp']), l)
        self.assertEqual(len(temporal['u_ydark']), l)
        self.assertEqual(len(temporal['s2_ydark']), l)

        del_obj(self.dataset, self.parser, self.loader, self.data)

    def test_data_errors(self):
        # Test that given an incomplete data dictionary, it will raise errors
        # if there is no dark data in temporal
        with self.assertRaises(ValueError):
            dat = {'width': 1, 'height': 1,
                   'temporal': {0: {0.1: None}},
                   'spatial': {0: {0.1: None}}}
            d = Data1288(dat)

        # if no dark data in spatial
        with self.assertRaises(ValueError):
            dat = {'width': 1, 'height': 1,
                   'temporal': {0: {0.0: {'sum': 0, 'pvar': 0},
                                    0.1: {'sum': 0, 'pvar': 0}}},
                   'spatial': {0: {0.1: {'sum': 0, 'pvar': 0}}}}
            d = Data1288(dat)

        # if there is no bright image for each dark
        with self.assertRaises(ValueError):
            dat = {'width': 1, 'height': 1,
                   'temporal': {0: {0.0: None}},
                   'spatial': {0: {0.0: None}}}
            d = Data1288(dat)

        # If there is no bright image for spatial
        with self.assertRaises(ValueError):
            dat = {'width': 1, 'height': 1,
                   'temporal': {0: {0.0: {'sum': 0, 'pvar': 0},
                                    0.1: {'sum': 0, 'pvar': 0}}},
                   'spatial': {0: {0.0: {'sum': 0, 'pvar': 0}}}}
            d = Data1288(dat)

        # If there is more than 1 exposure time with spatial data
        with self.assertRaises(ValueError):
            dat = {'width': 1, 'height': 1,
                   'temporal': {0: {0.0: {'sum': 0, 'pvar': 0},
                                    0.1: {'sum': 0, 'pvar': 0}}},
                   'spatial': {0: {0.0: {'sum': 0, 'pvar': 0},
                                   0.1: {'sum': 0, 'pvar': 0}},
                               1: {0.0: {'sum': 0, 'pvar': 0},
                                   0.1: {'sum': 0, 'pvar': 0}}}}
            d = Data1288(dat)



import unittest
from emva1288.process.parser import ParseEmvaDescriptorFile
from emva1288.process.loader import LoadImageData
from emva1288.process.data import Data1288
from emva1288.process.results import Results1288
from emva1288.camera.dataset_generator import DatasetGenerator
from emva1288.unittests.test_routines import del_obj
import numpy as np


def _init(pixel_area=0, **kwargs):
    # create dataset
    dataset = DatasetGenerator(**kwargs)

    # parse dataset
    parser = ParseEmvaDescriptorFile(dataset.descriptor_path)
    # load image data
    loader = LoadImageData(parser.images)
    # create data
    data = Data1288(loader.data)
    # Make results object
    px = pixel_area
    if pixel_area == 0:
        px = dataset.cam.pixel_area
    results = Results1288(data.data, pixel_area=px)
    return dataset, parser, loader, data, results


class TestResults(unittest.TestCase):
    _height = 50
    _width = 100
    _bit_depth = 8
    _L = 50
    _qe = 0.5
    _steps = 10
    _radiance_min = None
    _exposure_max = 5000000000
    _dark_current_ref = 30
    _exposure_fixed = 10000000
    _temperature = 20
    _temperature_ref = 20
    _K = 0.5
    _exposure_min = 50000
    _dsnu = np.zeros((_height, _width))
    _dsnu[0, :] += 5
    _prnu = np.ones((_height, _width))
    _prnu[-1, :] += 1.5

    def test_results_exposure_variation(self):
        dt, p, l, da, r = _init(height=self._height,
                                width=self._width,
                                bit_depth=self._bit_depth,
                                L=self._L,
                                qe=self._qe,
                                steps=self._steps,
                                radiance_min=self._radiance_min,
                                exposure_max=self._exposure_max,
                                exposure_min=self._exposure_min,
                                K=self._K,
                                dark_current_ref=self._dark_current_ref,
                                temperature=self._temperature,
                                temperature_ref=self._temperature_ref,
                                dsnu=self._dsnu,
                                prnu=self._prnu)

        self.dataset = dt
        self.parser = p
        self.loader = l
        self.data = da
        self.results = r

        ###########################
        # Test results properties #
        ###########################

        # test that quantification noise is really 1/12
        self.assertEqual(self.results.s2q, 1.0 / 12.0)

        # test that indexes are integers and in good range
        for attr in ('index_start', 'index_u_ysat', 'index_sensitivity_max',
                     'index_sensitivity_min', 'index_linearity_min',
                     'index_linearity_max'):
            value = getattr(self.results, attr)
            self.assertTrue(type(value) is int or type(value) is np.int64,
                            msg="%s is not an integer but should be!" % attr)
            self.assertTrue(value < len(self.data.data['temporal']['u_y']))
            self.assertTrue(value >= 0)
            self.assertTrue(attr in self.results.results.keys(),
                            msg="%s does not appear in the results!" % attr)

        # Test that EMVA values are float and positive
        for a in ('s2q', 'R', 'K', 'QE', 'sigma_y_dark', 'sigma_d', 'u_p_min',
                  'u_p_min_area', 'u_e_min', 'u_e_min_area', 'u_p_sat',
                  'u_p_sat_area', 'u_e_sat', 'SNR_max', 'DR', 'LE_min',
                  'LE_max', 'u_I_var', 'u_I_mean', 'sigma_2_y_stack',
                  'sigma_2_y_stack_dark', 's_2_y_measured', 's_2_y',
                  's_2_y_dark', 'DSNU1288', 'PRNU1288'):
            value = getattr(self.results, a)
            self.assertTrue(isinstance(value, float),
                            msg="%s is not a float but should be!" % a)
            if not a == 's2q':
                self.assertTrue(a in self.results.results,
                                msg="%s does not appear in the results!" % a)
            # except for linearity errors and dark currents,
            # everything always should be positive
            if (a not in ('LE_min', 'LE_max') and
               value is not np.nan):
                self.assertGreaterEqual(value, 0.0,
                                        msg="%s is negative but"
                                            " should be positive!" % a)

        ###############################################################
        # The following deltas are purely guesstimates and are prone to
        # errors in the future if they are not really significant
        ###############################################################

        # Test quantum efficiency is retrieved with a +/- 5% incertainty
        self.assertAlmostEqual(self._qe * 100, self.results.QE, delta=5.0,
                               msg="The difference between the expected QE and"
                                   "the retrieved one is greater than 5%!")

        # Test that overall system gain
        # is retrieved with a +/- 0.01 incertainty
        self.assertAlmostEqual(self.dataset.cam.K, self.results.K, delta=0.1,
                               msg="The difference between expected"
                                   "system gain"
                                   "and the retrieved one"
                                   "is greater than 0.01!")

        self.assertEqual(self.results.inverse_K(), 1 / self.results.K)

        # Test that responsivity is coherent with QE and system gain
        self.assertAlmostEqual(self.results.R,
                               self.results.QE * self.results.K / 100,
                               delta=0.001)  # division errors compensation

        # Test that dark current is actually retrieved from both methods
        self.assertAlmostEqual(self._dark_current_ref, self.results.u_I_mean,
                               delta=5.0,
                               msg="Dark current is not well retrieved from"
                                   " mean dark signal.")
        self.assertAlmostEqual(self._dark_current_ref, self.results.u_I_var,
                               delta=10.0,
                               msg="Dark current is not well retrieved from"
                                   " dark signal variance.")

        # Test that u_e_sat_area = u_e_sat / area
        self.assertAlmostEqual(self.results.u_e_sat_area,
                               self.results.u_e_sat /
                               self.dataset.cam.pixel_area,
                               delta=0.01)

        # Test that SNR_max is sqrt of u_e_sat
        self.assertAlmostEqual(self.results.SNR_max,
                               np.sqrt(self.results.u_e_sat),
                               delta=0.01)

        # Test that SNR_max_db is 20log_10(SNR_max)
        self.assertAlmostEqual(self.results.SNR_max_dB(),
                               20 * np.log10(self.results.SNR_max),
                               delta=0.01)

        # Test that SNR_max_bit is log_2(SNR_max)
        self.assertAlmostEqual(self.results.SNR_max_bit(),
                               np.log2(self.results.SNR_max),
                               delta=0.01)

        # Test that SNR_max inverse is 100 / SNR_max
        self.assertAlmostEqual(self.results.inverse_SNR_max(),
                               100 / self.results.SNR_max,
                               delta=0.01)

        # Test that DR is u_p_sat / u_p_min
        self.assertAlmostEqual(self.results.DR,
                               self.results.u_p_sat / self.results.u_p_min,
                               delta=0.01)

        # Test that DR_dB is 20log_10(DR)
        self.assertAlmostEqual(self.results.DR_dB(),
                               20 * np.log10(self.results.DR),
                               delta=0.01)

        # Test that DR_bit is log_2(DR)
        self.assertAlmostEqual(self.results.DR_bit(),
                               np.log2(self.results.DR),
                               delta=0.01)

        # Test that DSNU is sqrt(s2_ydark) / gain
        self.assertAlmostEqual(self.results.DSNU1288,
                               np.sqrt(self.results.s_2_y_dark) /
                               self.results.K,
                               delta=0.01)

        # Test that DSNU in DN is DSNU * K
        self.assertAlmostEqual(self.results.DSNU1288_DN(),
                               self.results.DSNU1288 * self.results.K,
                               delta=0.01)

        # Test that PRNU is the same as defined in EMVA1288 standard
        self.assertAlmostEqual(self.results.PRNU1288,
                               np.sqrt(self.results.s_2_y -
                                       self.results.s_2_y_dark) * 100 /
                               (self.data.data['spatial']['avg_mean'] -
                                self.data.data['spatial']['avg_mean_dark']))

        # Test that histograms contains relevant keys and are numpy arrays
        hists = ('histogram_PRNU', 'histogram_PRNU_accumulated',
                 'histogram_DSNU', 'histogram_DSNU_accumulated')
        keys = ('bins', 'model', 'values')
        for hist in hists:
            h = getattr(self.results, hist)
            for key in keys:
                self.assertTrue(key in h.keys())
                self.assertTrue(isinstance(h[key], np.ndarray))

        # delete objects
        del_obj(self.dataset, self.parser, self.loader, self.data,
                self.results)

    def test_results_current_variation(self):
        dt, p, l, da, r = _init(height=self._height,
                                width=self._width,
                                bit_depth=self._bit_depth,
                                L=self._L,
                                steps=self._steps,
                                dark_current_ref=self._dark_current_ref,
                                exposure_fixed=self._exposure_fixed,
                                radiance_min=self._radiance_min,
                                exposure_max=self._exposure_max,
                                exposure_min=self._exposure_min,
                                temperature=self._temperature,
                                temperature_ref=self._temperature_ref,
                                K=self._K,
                                dsnu=self._dsnu,
                                prnu=self._prnu)
        self.dataset = dt
        self.parser = p
        self.loader = l
        self.data = da
        self.results = r

        data = self.data.data
        # Test that s_ydark is not a fit because only 1 texp
        self.assertAlmostEqual(self.results.sigma_y_dark,
                               np.sqrt(data['temporal']['s2_ydark'][0]),
                               delta=0.1)

        del_obj(self.dataset, self.parser, self.loader, self.data,
                self.results)

    def test_results_without_pixel_area(self):
        dt, p, l, da, r = _init(pixel_area=None,
                                height=self._height,
                                width=self._width,
                                bit_depth=self._bit_depth,
                                L=self._L,
                                steps=self._steps,
                                dark_current_ref=self._dark_current_ref,
                                exposure_max=self._exposure_max,
                                exposure_min=self._exposure_min,
                                temperature=self._temperature,
                                temperature_ref=self._temperature_ref,
                                K=self._K,
                                dsnu=self._dsnu,
                                prnu=self._prnu)
        self.dataset = dt
        self.parser = p
        self.loader = l
        self.data = da
        self.results = r

        # Test relevant properties are None
        self.assertIs(self.results.u_p_min_area, None)
        self.assertIs(self.results.u_e_min_area, None)
        self.assertIs(self.results.u_p_sat_area, None)
        self.assertIs(self.results.u_e_sat_area, None)

        del_obj(self.dataset, self.parser, self.loader, self.data,
                self.results)

    def test_nans(self):
        # Test that less than 2 texp will yield a NaN for u_I_mean
        data = {'temporal': {'texp': [0, 1]},
                'spatial': {}}
        r = Results1288(data)
        self.assertIs(r.u_I_mean, np.nan)

        # Test that a negative slope for t vs s2_ydark will yield Nan for
        # u_I_var
        data['temporal']['s2_ydark'] = [1, 0]
        r = Results1288(data)
        self.assertIs(r.u_I_var, np.nan)

        # Test that a negative s2y_dark will yield a Nan for DSNU1288
        data['spatial'] = {'avg_var_dark': 0.,
                           'var_mean_dark': 1.,
                           'L_dark': 3}
        r = Results1288(data)
        self.assertIs(r.DSNU1288, np.nan)
        self.assertIs(r.DSNU1288_DN(), np.nan)

        del r



import unittest
from emva1288.camera.camera import Camera
import numpy as np
from emva1288.camera import routines


class CameraTestCase(unittest.TestCase):
    def setUp(self):
        self.cam = Camera()

    def tearDown(self):
        del self.cam

    def test_img(self):
        img = self.cam.grab(0)
        self.assertEqual((self.cam.height, self.cam.width), np.shape(img))

    def test_radiance(self):
        img1 = self.cam.grab(0)
        img2 = self.cam.grab(self.cam.get_radiance_for(mean=250))
        self.assertLess(img1.mean(), img2.mean())


class CameraTestBayer(unittest.TestCase):
    def test_get_bayer(self):
        # Init the parameters
        h, w = [7, 5]

        transmition_pixel_1 = 1
        transmition_pixel_2 = 2
        transmition_pixel_3 = 3
        transmition_pixel_4 = 4
        b_layer = routines.get_bayer_filter(transmition_pixel_1,
                                            transmition_pixel_2,
                                            transmition_pixel_3,
                                            transmition_pixel_4, w, h)
        # Supposed Results
        lines = [1, 2, 1, 2, 1]
        columns = [1, 3, 1, 3, 1, 3, 1]
        # Test to see if the layer come right
        self.assertEqual(lines, b_layer[0].tolist())
        self.assertEqual(columns, b_layer[:, 0].tolist())

    def test_bayer_layer(self):
        # Init the parameters
        h, w = [480, 640]
        transmition_red = 0.15
        transmition_blue = 0.02
        transmition_green = 1.
        b_layer = routines.get_bayer_filter(transmition_green,
                                            transmition_red,
                                            transmition_blue,
                                            transmition_green, w, h)
        # Test if the b_layer have the same shape than what we give it
        self.assertEqual((h, w), b_layer.shape)
        # Set the camera for testing the layer
        cam = Camera(width=w, height=h, radiance_factor=b_layer)
        target = cam.img_max / 2
        radiance = cam.get_radiance_for(mean=target)
        img = cam.grab(radiance)
        green_filter = routines.get_bayer_filter(0, 1, 1, 0, w, h)
        blue_filter = routines.get_bayer_filter(1, 1, 0, 1, w, h)
        red_filter = routines.get_bayer_filter(1, 0, 1, 1, w, h)
        # Test if the mean of the green it's 100% of the target +/- 5
        self.assertAlmostEqual(np.ma.masked_array(
            img,
            mask=green_filter).mean(),
            target, delta=5.0,
            msg="green not in range")
        # Test if the mean of the red it's 15% of the target +/- 5
        self.assertAlmostEqual(np.ma.masked_array(
            img,
            mask=red_filter).mean(),
            target*transmition_red, delta=5.0,
            msg="red not in range")
        # Test if the mean of the blue it's 2% of the target +/- 5
        self.assertAlmostEqual(np.ma.masked_array(
            img,
            mask=blue_filter).mean(),
            target*transmition_blue, delta=5.0,
            msg="blue not in range")


class CameraTestPrnuDsnu(unittest.TestCase):
    def test_prnu(self):
        # Init the parameters
        h, w = [480, 640]
        rep = 200
        value8 = 3
        # create the pattern of the prnu
        prnu_array = np.ones((8))
        prnu_array[-1] = value8
        prnu = routines.get_tile(prnu_array, h, w)
        # Set the camera for testing the prnu
        cam = Camera(width=w, height=h, prnu=prnu)
        var = np.sqrt(cam._sigma2_dark_0)
        target = cam.img_max / 2
        # The target (top_target) is the multiplication of the target
        # (what we expect without prnu) and the value8(prnu). We can do it
        # because if we look at the _u_e function in emva1288.camera.camera
        # the prnu affect the QE with a multiplication. So if we just
        # multiplied the target by the prnu it's the same thing.
        # But this value can go over the maximal value for one pixel, this
        # is why we use the min() function to take the maximal value than the
        # camera can take.
        top_target = min(target * value8, cam.img_max)
        radiance = cam.get_radiance_for(mean=target)
        img = cam.grab(radiance)
        # create the mask
        prnu_mask = np.zeros((8))
        prnu_mask[-1] = 1
        prnu_mask_resize = routines.get_tile(prnu_mask, h, w)
        prnu_non_mask = np.ones((8))
        prnu_non_mask[-1] = 0
        prnu_non_mask_resize = routines.get_tile(prnu_non_mask, h, w)
        # Test if the mean it's 100% of the target +/- variance
        self.assertAlmostEqual(np.ma.masked_array(
            img,
            mask=prnu_mask_resize).mean(),
            target, delta=var,
            msg="values are not in range")
        # Test if the mean of the 8th value it's value8
        # multiplied be the target +/- variance
        self.assertAlmostEqual(np.ma.masked_array(
            img,
            mask=prnu_non_mask_resize).mean(),
            top_target, delta=var,
            msg="8th value it's not in range")

    def test_dsnu(self):
        # Init the parameters
        h, w = [480, 640]
        value8 = 5
        rep = 200
        # create the pattern of the dsnu
        dsnu_array = np.ones((8))
        dsnu_array[-1] = value8
        dsnu = routines.get_tile(dsnu_array, h, w)
        # Set the camera for testing the dsnu
        cam = Camera(width=w, height=h, dsnu=dsnu)
        var = np.sqrt(cam._sigma2_dark_0)
        # The target is the number of electrons who are not affected
        # by the dsnu. To resume, we suppose to observe is a combinaison of
        # electrons from the dark signal and the temperature. The total need
        # to be multiplied by the gain of the system (K).
        # for more eplination see the grab function in emva1288.camera.camera
        target = cam.K * (cam._dark_signal_0 + cam._u_therm())
        # Here the target (top_target) is the part who is affected by
        # the dsnu. Physicaly, the same phenomen append but this time the
        # dark signal is NonUniform so thw value who represent the dsnu is
        # added to the dark signal befor the multiplication of the gain.
        top_target = cam.K * (cam._dark_signal_0 + cam._u_therm() + value8)
        img = cam.grab(0)
        # create the mask
        dsnu_mask = np.zeros((8))
        dsnu_mask[-1] = 1
        dsnu_mask_resize = routines.get_tile(dsnu_mask, h, w)
        dsnu_non_mask = np.ones((8))
        dsnu_non_mask[-1] = 0
        dsnu_non_mask_resize = routines.get_tile(dsnu_non_mask, h, w)
        # Test if the mean it's 100% of the target +/- variance
        self.assertAlmostEqual(np.ma.masked_array(
            img,
            mask=dsnu_mask_resize).mean(),
            target, delta=var,
            msg="values are not in range")
        # Test if the mean of the 8th value it's value8
        # multiplied be the target +/- variance
        self.assertAlmostEqual(np.ma.masked_array(
            img,
            mask=dsnu_non_mask_resize).mean(),
            top_target, delta=var,
            msg="8th value it's not in range")


class CameraTestRoutines(unittest.TestCase):
    def test_get_tile(self):
        # TEST 1D
        # Init the parameters
        h, w = [1, 24]
        dim1 = np.zeros((8))
        # Supposed Results
        res_array = routines.get_tile(dim1, h, w)
        res_dim1 = np.zeros((24))
        # Test to see if the layer come right
        # shape[0] will give us the number in (width,)
        self.assertEqual(w, res_array.shape[0])
        self.assertEqual(res_dim1.tolist(), res_array.tolist())

        # TEST 2D
        # Init the parameters
        h, w = [5, 7]
        dim2 = np.zeros((3))
        # Supposed Results
        res_array = routines.get_tile(dim2, h, w)
        res_dim2 = np.zeros((5, 7))
        # Test to see if the layer come right
        self.assertEqual((h, w), res_array.shape)
        self.assertEqual(res_dim2.tolist(), res_array.tolist())



# This module contains a generator that generates a descriptor file and the
# corresponding images using the implemented camera.

from emva1288.camera.camera import Camera as Cam
from emva1288.camera.points_generator import PointsGenerator
from collections import OrderedDict
import numpy as np
import tempfile
import os
from PIL import Image


def _get_emva_gain(cam):
    """Find the gain to satisfy EMVA1288 requirements"""
    gini = cam.K
    # Find gain with a minum temporal noise of 0.5DN
    g = cam.Ks[0]
    for gain in cam.Ks:
        cam.K = gain
        g = gain
        img1 = cam.grab(0).astype(np.int64)
        img2 = cam.grab(0).astype(np.int64)
        if (img1 - img2).std() > 0.5:
            break
    cam.K = gini
    return g


def _get_emva_blackoffset(cam):
    """Find the blackoffset to satifsfy EMVA1288 requirements"""
    bini = cam.blackoffset
    # Find black offset with a maximum of 0.5% of values at Zero
    bo = cam.blackoffsets[0]
    pixels = cam.width * cam.height
    for i in cam.blackoffsets:
        cam.blackoffset = i
        img = cam.grab(0)
        bo = i
        if np.count_nonzero(img) > pixels * .995:
            break
    cam.blackoffset = bini
    return bo


class DatasetGenerator:
    """Dataset generator.

    Creates a descriptor file and the corresponding linked images for a
    a exposure variant test example according to the emva1288 standart.
    The images are created using the implemented camera in the emva module.
    """

    def __init__(self,
                 steps=100,
                 L=50,
                 version='3.0',
                 image_format='png',  # best memory consumption
                 outdir=None,  # directory where to save the dataset
                 radiance_min=None,
                 radiance_max=None,
                 exposure_fixed=None,
                 **kwargs
                 ):
        """Dataset generator init method.

        The generator uses a
        :class:`~emva1288.camera.points_generator.PointsGenerator`
        object to create the operation points. It then grabs the images
        for these points using a
        :class:`~emva1288.camera.camera.Camera` simulator object.
        The camera is intialized according to the
        given kwargs. Then, after getting the test points,
        it :meth:`makes <run_test>` the images
        with it by changing its
        exposure time, or the radiation and :meth:`saves <save_images>`
        the images and the descriptor file.

        Parameters
        ----------
        L : int, optional
            The number of image taken during a spatial test point.
        version : str, optional
                  Data version to add in descriptor file.
        image_format : str, optional
                       The image's format when they are saved.
        outdir : str, optional
                 The output directory where the descriptor file and the images
                 will be saved. If None, it will create a tempory directory
                 that will be deleted (and its contents) when the dataset
                 generator object is deleted.
        radiance_min : float, optional
                       Same as in
                       :class:`~emva1288.camera.points_generator.PointsGenerator`.
        radiance_max : float, optional
                       Same as in
                       :class:`~emva1288.camera.points_generator.PointsGenerator`.
        exposure_fixed : float, optional
                         Same as in
                         :class:`~emva1288.camera.points_generator.PointsGenerator`.
        kwargs : All other kwargs are passed to the camera.
        """
        self._steps = steps  # number of points to take
        self.cam = Cam(**kwargs)
        # set the camera parameters for the test
        self.cam.exposure = self.cam.exposure_min

        # If no blackoffset/gain are specified find them according to standard
        if 'blackoffset' not in kwargs:
            self.cam.blackoffset = _get_emva_blackoffset(self.cam)
        if 'K' not in kwargs:
            self.cam.K = _get_emva_gain(self.cam)

        # create test points
        points = PointsGenerator(self.cam,
                                 radiance_min=radiance_min,
                                 radiance_max=radiance_max,
                                 exposure_fixed=exposure_fixed,
                                 steps=self._steps)
        self._points = points.points

        self._L = L  # number of images to take for a spatial test
        self._version = version  # data version
        # store image format
        self._image_format = image_format

        # images will be saved one at a time during the generation into outdir
        self.outdir = outdir
        # create temporary directory to store the dataset
        if outdir is None:
            self.tempdir = tempfile.TemporaryDirectory()
            self.outdir = self.tempdir.name
        # create dir where images will be saved
        os.makedirs(os.path.join(self.outdir, 'images'))

        # run test
        self._descriptor_path = self.run_test()

    @property
    def points(self):
        """The test points suite."""
        return self._points

    @property
    def descriptor_path(self):
        """The absolute path to the descriptor file."""
        return self._descriptor_path

    def _is_point_spatial_test(self, i):
        """Check if a point index should be a spatial test.

        Spatial points are done at midpoint of bright and dark series.
        """
        v = self._steps // 2
        if i in (v, self._steps + v):
            return True
        return False

    def _get_descriptor_line(self, exposure, radiance):
        """Create the line introducing a test point images in descriptor."""
        if radiance == 0.0:
            # dark image
            return "d %.1f" % exposure
        # bright image
        # round photons count to three decimals
        return "b %.1f %.3f" % (exposure,
                                round(self.cam.get_photons(radiance), 3))

    def _get_image_names(self, number, L):
        """Create an image filename."""
        names = []
        for l in range(L):
            names.append("img_%04d.%s" % (number, self._image_format))
            number += 1
        return names, number

    def _get_imgs(self, radiance, L):
        """Create a list of image from the given radiances.
        """
        # computes an array of dict whose keys are the name of the file
        # and the data of the image to save
        imgs = []
        for l in range(L):
            imgs.append(self.cam.grab(radiance))
        return imgs

    def run_test(self):
        """Run the test points, save the images and generate descriptor."""
        descriptor_text = OrderedDict()
        image_number = 0
        # descriptor file path
        path = os.path.join(self.outdir, "EMVA1288descriptor.txt")
        # open descriptor file to write images in it
        with open(path, "w") as f:
            # write version
            f.write("v %s\n" % self._version)
            # wtite camera's properties
            f.write("n %i %i %i\n" % (self.cam.bit_depth,
                                      self.cam.width,
                                      self.cam.height))
            for kind in ('temporal', 'spatial'):

                # number of image to take
                L = 2
                if kind == 'spatial':
                    L = self._L
                for texp, radiances in self.points[kind].items():
                    # set camera
                    self.cam.exposure = texp
                    # Grab all images for these radiances
                    for radiance in radiances:
                        # Get descriptor line introducting the images
                        f.write("%s\n"
                                % self._get_descriptor_line(texp, radiance))
                        for l in range(L):
                            # grab
                            img = self.cam.grab(radiance)
                            # write the name in descriptor
                            name = "image%i.%s" % (image_number,
                                                   self._image_format)
                            f.write("i images\\%s\n" % name)
                            image_number += 1

                            # save image
                            self.save_image(img, name)

        # return descriptor path
        return path

    def save_image(self, img, name):
        """Save the image.
        """
        # save the images contained in d
        dtype = np.uint32
        mode = 'I'
        if self.cam.bit_depth <= 8:
            # 8 bit images have special format for PIL
            dtype = np.uint8
            mode = 'L'
        im = Image.fromarray(img.astype(dtype), mode)
        path = os.path.join(self.outdir, 'images', name)
        # (filename already contains image format)
        im.save(path)
        # erase image from dict to reduce memory consuption



from setuptools import setup
import os

# Get the version from versioneer
import versioneer
__version__ = versioneer.get_version()

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.rst')) as f:
    long_description = f.read()

# On read the docs we don't install anything
on_rtd = os.environ.get('READTHEDOCS') == 'True'
if on_rtd:
    install_requires = []
else:
    install_requires = ['matplotlib',
                        'numpy',
                        'pillow',
                        'lxml',
                        'scipy',
                        'jinja2']

setup(name='emva1288',
      packages=['emva1288',
                'emva1288.report',
                'emva1288.camera',
                'emva1288.process',
                'emva1288.unittests'],
      version=__version__,
      description='EMVA1288 reference implementation',
      long_description=long_description,
      author='Federico Ariza',
      author_email='ariza.federico@gmail.com',
      url='https://github.com/EMVA1288/emva1288',
      keywords=['sensors', 'cameras'],
      classifiers=['Programming Language :: Python :: 2',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.4'],
      install_requires=install_requires,
      package_data={'emva1288': ['report/files/*', 'report/templates/*',
                                 'examples/*']}
      )



import numpy as np
from emva1288.camera import Camera
from emva1288.process.routines import get_int_imgs
from emva1288 import process


def get_emva_blackoffset(cam):
    """Find the blackoffset to satifsfy EMVA1288 requirements"""
    bini = cam.blackoffset
    # Find black offset with a maximum of 0.5% of values at Zero
    bo = cam.blackoffsets[0]
    pixels = cam.width * cam.height
    for i in cam.blackoffsets:
        cam.blackoffset = i
        img = cam.grab(0)
        bo = i
        if np.count_nonzero(img) > pixels * .995:
            break
    cam.blackoffset = bini
    return bo


def get_emva_gain(cam):
    """Find the gain to satisfy EMVA1288 requirements"""
    gini = cam.K
    # Find gain with a minum temporal noise of 0.5DN
    g = cam.Ks[0]
    for gain in cam.Ks:
        cam.K = gain
        g = gain
        img1 = cam.grab(0).astype(np.int64)
        img2 = cam.grab(0).astype(np.int64)
        if (img1 - img2).std() > 0.5:
            break
    cam.K = gini
    return g


def get_temporal(cam, radiance):
    img1 = cam.grab(radiance)
    img2 = cam.grab(radiance)
    imgs = get_int_imgs((img1, img2))
    return {'sum': np.sum(imgs['sum']), 'pvar': np.sum(imgs['pvar'])}


def get_spatial(cam, radiance, L=50):
    imgs = []
    for i in range(L):
        imgs.append(cam.grab(radiance))
    return get_int_imgs(imgs)


data = {'temporal': {},
        'spatial': {},
        'width': None, 'height': None}


# Intialize the camera, here we can specify different image size
# or any other parameter that Camera allows
c = Camera(bit_depth=10,
           width=100,
           height=50)

# Fill the information
data['width'] = c.width
data['height'] = c.height

# Maximum exposure for test
exposure_max = 9000000

# Find the camera parameters for the test
c.exposure = exposure_max
c.blackoffset = get_emva_blackoffset(c)
c.K = get_emva_gain(c)

# Find the radiance that will saturate the camera at our maximum exposure time
saturation_radiance = c.get_radiance_for()

# Initialize the exposure for the spatial measure
exposure_spatial = None

# Loop through the exposures
for exposure in np.linspace(c.exposure_min, exposure_max, 100):
    c.exposure = exposure
    data['temporal'][exposure] = {}

    # For each exposure, take to measurements (bright, dark)
    for radiance in (saturation_radiance, 0.0):
        photons = c.get_photons(radiance)

        # Get the temporal data
        data['temporal'][exposure].setdefault(photons, {})
        data['temporal'][exposure][photons] = get_temporal(c, radiance)

        # Check if we are at the middle of the range, to set the spatial exp
        img = c.grab(radiance)
        if not exposure_spatial and (img.mean() > c.img_max / 2.):
            exposure_spatial = exposure

        # Get the spatial data
        if exposure_spatial == exposure:
            data['spatial'].setdefault(exposure, {})
            data['spatial'][exposure].setdefault(photons, {})
            data['spatial'][exposure][photons] = get_spatial(c, radiance)


# Process the collected data
dat = process.Data1288(data)
res = process.Results1288(dat.data, pixel_area=c.pixel_area)
res.print_results()
plot = process.Plotting1288(res)
plot.plot()
